open Items
open Table
open Lattice
open SolverEq
open SolverWiden
open SolverCon
open SolverConWiden


module Graph (X: Item) (D: Lattice) =
struct
   module T = Table (X)
   type 'a graph =  (('a -> 'a) * X.item) list T.table 

   let graph_of_list lst = let g = T.make_table 101
			in  let do_edge (x,f,y) = match T.sub g y
				with None -> T.add g y [(f,x)]
				| Some fx -> T.update g y ((f,x)::fx)
			in  let _ = List.iter do_edge lst
			in  g

   let eq g (start,d) y = let transfer (f,x) = fun sigma -> f (sigma x)
		       in let constr = match T.sub g y
				with None -> []
				| Some edges -> List.map transfer edges
		       in if y = start then (fun _ -> d) :: constr
		          else constr

   let eq_of_list edges = eq (graph_of_list edges)

   let eq_of_constr constr x sigma = let doit a f = D.cup a (f sigma)
				      in List.fold_left doit (D.bot()) (constr x)

   module SE = SolverEq (X) (D)

   let solve_eq g (start,d) lst = let constr = eq g (start,d)
			           in SE.solve (eq_of_constr constr) lst

   let solve_list_eq edges = solve_eq (graph_of_list edges)

   module SW = SolverWiden (X) (D)

   let solve_widen g (start,d) lst = let constr = eq g (start,d)
			              in SW.solve (eq_of_constr constr) lst

   let solve_list_widen edges = solve_widen (graph_of_list edges)

   module SC = SolverCon (X) (D)

   let solve g (start,d) lst = let constr = eq g (start,d)
			        in SC.solve constr lst

   let solve_list edges = solve (graph_of_list edges)

   module SCW = SolverConWiden (X) (D)

   let solve_con_widen g (start,d) lst = let constr = eq g (start,d)
			                  in SCW.solve constr lst

   let solve_list_con_widen edges = solve_con_widen (graph_of_list edges)

end

(*
module P = Power(StringItem)
module G = Graph (IntItem)(P)

let a1 = P.from_list ["a"; "b"]
let a2 = P.from_list ["b"; "c"]
let a3 = P.from_list ["c"; "d"]

let f1 x = P.cap x a1
let f2 x = P.cup x a2
let f3 x = P.cup (P.cap x a3) a1

let edges = [(1,f1,2); (2,f2,3); (3,f3,2)]
let (start,d) = (1,P.bot())

let result = let doit (x,d) = (x,P.to_list d)
	     in List.map doit (G.S.T.to_list (
			       G.solve_list edges (start,d) [3]))

*)
