open Items
open Table
open Pool
open Lattice
open SolverCon
open SolverConWiden

type ('a,'p) lab = Basic of ('a -> 'a) 
		 | Call of (('a -> 'a) * ('a -> 'p list) * ('a -> 'a -> 'a))

module Graph2 (X: Item) (P: Item) (D: Lattice) =
struct
   module I = PairItem (P)(OptionItem(X))
   module I0 = PairItem (P)(X)
   module V = PairItem (I)(D)
   module T0 = Table (I0)
   module TP = Table (P)
   module S0 = Pool (I0)

   type 'a proc_graph =  (I0.item * ('a,P.item) lab) list T0.table 
   type 'a graph = (X.item * X.item list * 'a proc_graph) TP.table

   let proc_graph_of_list g (p,list) = 
			    let do_edge (x,f,y) = match T0.sub g (p,y)
				with None -> T0.add g (p,y) [((p,x),f)]
				| Some fx -> T0.update g (p,y) (((p,x),f)::fx)
			    in  List.iter do_edge list

   let graph_of_list list = let gp = TP.make_table 51
			in  let g  = T0.make_table 128
			in  let first = S0.make_set 51
			in  let do_proc (p,(x,res,edges)) = 
					TP.update gp p res;
				   	proc_graph_of_list g (p,edges);
					S0.insert first (p,x)
			in  let _ = List.iter do_proc list
			in (g,gp,first)

   type var = V.item
   
   let eq (g,gp,first) y =  let transfer d ((p,x),lab) = let var = ((p, Some x), d)
			     in match lab
			 	with Basic f -> fun sigma -> f (sigma var)
				| Call (entry,call,comb)  -> let do_call sigma a f = 
					    D.cup a (comb d (sigma ((f, None), entry d)))
					in fun sigma -> List.fold_left (do_call sigma) 
								       (D.bot()) 
								       (call d)
		            in match y
			 	with ((p,Some y),a) -> let edges = (match T0.sub g (p,y)
							     with None -> []
							   | Some list -> list
								)
				             in let list = List.map (transfer a) edges
					     in if S0.has_item first (p,y) then 
					      		(fun _ -> a) :: list
					   	else list
				| ((p,None),a) -> (match TP.sub gp p
					  with None  -> []
					  | Some res -> List.map (fun x -> 
						fun sigma -> sigma ((p,Some x),a)) res
						)

   let eq_of_list edges x = eq (graph_of_list edges) x

   module S = SolverCon (V) (D)

   let solve (g,gp,first) (main,d) = let eq = eq (g,gp,first) 
			       in let x = ((main,None), d)
			        in S.solve eq [x]

   let solve_list edges = solve (graph_of_list edges)

   module SW = SolverConWiden (V) (D)

   let solve_widen (g,gp,first) (main,d) = let constr = eq (g,gp,first)
                               in let x = ((main,None), d)
                                in SW.solve constr [x]

   let solve_list_widen edges = solve_widen (graph_of_list edges)

end

module P = Power(StringItem)
module G = Graph2 (IntItem)(StringItem)(P)

let a1 = P.from_list ["a"; "b"]
let a2 = P.from_list ["b"; "c"]
let a3 = P.from_list ["c"; "d"]

let f1 x = P.cap x a1
let f2 x = P.cup x a2
let f3 x = P.cup (P.cap x a3) a1

let edges : (string * (int * (int list) * ((int * (P.d, string) lab * int) list))) list 
	  = [("main",(1,[3],[(1,Basic f1,2); (2,Basic f2,3); (3,Basic f3,2)]))]
(*
	  = [("main",(1,[2],[(1,Basic f1,2)]))]
*)

let result = let doit (x,d) = (x,P.to_list_option d)
	     in List.map doit (G.S.T.to_list (
			       G.solve_list edges ("main",P.top())))


