open Items
open Lattice
open Table
open Pool
   
module SolverCon (X: Item) (D: Lattice) = 
struct
   module T = Table (X)
   module P = Pool  (X)

   let rec iter f = function
	  []    -> ()
	| [x]   -> f x
	| x::xs -> (f x; iter f xs)

   type rhs    = (X.item -> D.d) -> D.d
   type constr =  X.item * rhs

   let solve constr list = 

   	let sigma : D.d T.table = T.make_table 128
   	in let get x = match T.sub sigma x
                with None -> D.bot ()
		| Some d  -> d
   	in let set x d = T.update sigma x d

        in let to_do : rhs list T.table = T.make_table 128
	in let extract_todo x = let result = T.sub to_do x
			     in let _ = T.update to_do x []
			     in    match result 
				   with None -> constr x 
				   | Some fs -> fs
  	in let add_todo x f = match T.sub to_do x
			      with None -> T.add to_do x [f]
			      | Some fs -> T.update to_do x (f::fs)

   	in let infl : constr list T.table = T.make_table 128
   	in let add_infl y (x,f) = match T.sub infl y
                  with None -> T.update infl y [(x,f)]
                  | Some xs -> T.update infl y ((x,f)::xs)
   	in let extract_infl x = let w = T.sub infl x
		     in let _ = T.update infl x []
		     in let fs = match w
		  		with None -> []
		  		| Some xs -> xs
		     in let doit (x,f) = (add_todo x f; x)
		     in Srt.rem_dup (List.map doit fs)

   	in let rec eval (x,f) y =  solve y; 
		       	add_infl y (x,f); 
		       	get y

   	   and solve x = let old_val = get x
		      in let combine d f = let d = D.cup d (f (eval (x,f)))
					in let _ = set x d
					in d
		      in if D.leq (List.fold_left combine old_val 
					(extract_todo x)) old_val then ()
			 else iter solve (extract_infl x)

	in iter solve list; 
	   sigma
		
end
	
				      
     
