open Items
open Lattice
open Table
open Pool
   
module SolverConWiden (X: Item) (D: Lattice) = 
struct
   module T = Table (X)
   module P = Pool  (X)

   let rec iter f = function
	  []    -> ()
	| [x]   -> f x
	| x::xs -> (f x; iter f xs)

   let sigma : D.d T.table = T.make_table 128
   let get x = match T.sub sigma x
                with None -> D.bot ()
		| Some d  -> d
   let set x d = T.update sigma x d


   type rhs    = (X.item -> D.d) -> D.d
   type constr =  X.item * rhs

   let solve constr list = 

           let to_do : rhs list T.table = T.make_table 128
        in let extract_todo x = let result = T.sub to_do x
                             in let _ = T.update to_do x []
                             in    match result
                                   with None -> constr x
                                   | Some fs -> fs
        in let add_todo x f = match T.sub to_do x
                              with None -> T.add to_do x [f]
                              | Some fs -> T.update to_do x (f::fs)

        in let infl_w : constr list T.table = T.make_table 128
        in let add_infl_w y (x,f) = match T.sub infl_w y
                  with None -> T.update infl_w y [(x,f)]
                  | Some xs -> T.update infl_w y ((x,f)::xs)
        in let extract_infl_w x = let w = T.sub infl_w x
                     in let _ = T.update infl_w x []
                     in let fs = match w
                                with None -> []
                                | Some xs -> xs
                     in let doit (x,f) = (add_todo x f; x)
                     in Srt.rem_dup (List.map doit fs)

        in let rec eval_w (x,f) y =  solve_w y;
                        add_infl_w y (x,f);
                        get y

           and solve_w x = let old_val = get x
                      in let combine d f = D.cup d (f (eval_w (x,f)))
                      in let new_val = List.fold_left combine old_val (extract_todo x)
		      in let new_val = D.widen old_val new_val
		      in if D.leq new_val old_val then ()
                         else (set x new_val;
			       iter solve_w (extract_infl_w x)
			      )

   	in let stable : P.set = P.make_set 128
	in let is_stable x = P.has_item stable x 
  	in let add_stable x = P.add_item stable x
	in let rem_stable x = P.rem_item stable x

   	in let infl : X.item list T.table = T.make_table 128
   	in let add_infl y x = match T.sub infl y
                  with None -> T.update infl y [x]
                  | Some xs -> T.update infl y (x::xs)
  	in let extract_infl x = let w = T.sub infl x
			     in let _ = T.update infl x []
			     in match w
		  		with None -> []
		  		| Some xs -> xs

	in let eq x sigma = let doit a f = D.cup a (f sigma)
			    in List.fold_left doit (D.bot()) (constr x)
 
	in let rec eval x y = solve y;
                        add_infl y x;
                        get y

	   and solve_n x = if is_stable x then ()
                    else let _ = add_stable x
                      in let old_val = get x
                      in let new_val = D.narrow old_val (eq x (eval x))
		      in if D.leq old_val new_val then ()
                         else let _ = set x new_val
                           in let work = extract_infl x
                           in let _ = iter rem_stable work
                           in iter solve_n work
                         
	   and solve x = match T.sub sigma x
			 with None -> (solve_w x; solve_n x)
			 | Some _ -> solve_n x

	in iter solve list; 
	   sigma
		
end
	
(*
module I = IntLattice
module Test = SolverConWiden (StringItem)(I) 

let eq = function
	 "a" -> fun sigma -> I.cup (sigma "a") (sigma "b")
       | "b" -> fun sigma -> sigma "d" + 1
       | "c" -> fun sigma -> I.cap 0 (sigma "b")
       | _   -> fun sigma -> 1
     
let result = Test.T.to_list (Test.solve eq ["a"])
*)

