open Items
open Lattice
open Table
open Pool
   
module SolverEq (X: Item) (D: Lattice) = 
struct
   module T = Table (X)
   module P = Pool  (X)

   let rec iter f = function
	  []    -> ()
	| [x]   -> f x
	| x::xs -> (f x; iter f xs)


   let solve eq list = 

   	let sigma : D.d T.table = T.make_table 128
   	in let get x = match T.sub sigma x
                with None -> D.bot ()
		| Some d  -> d
   	in let set x d = T.update sigma x d

   	in let stable : P.set = P.make_set 128
	in let is_stable x = P.has_item stable x 
  	in let add_stable x = P.add_item stable x
	in let rem_stable x = P.rem_item stable x

   	in let infl : X.item list T.table = T.make_table 128
   	in let add_infl y x = match T.sub infl y
                  with None -> T.update infl y [x]
                  | Some xs -> T.update infl y (x::xs)
  	in let extract_infl x = let w = T.sub infl x
			     in let _ = T.update infl x []
			     in match w
		  		with None -> []
		  		| Some xs -> xs

   	in let rec eval x y =  solve y; 
		       	add_infl y x; 
		       	get y

           and solve x = if is_stable x then ()
                    else let _ = add_stable x
                      in let old_val = get x
                      in let new_val = D.cup old_val (eq x (eval x))
                      in if D.leq new_val old_val then ()
                         else let _ = set x new_val
                           in let work = extract_infl x
                           in let _ = iter rem_stable work
                           in iter solve work

	in iter solve list; 
	   sigma
		
end
	
(*
module I = IntLattice
module Test = SolverEq (StringItem)(I) 

let eq = function
	 "a" -> fun sigma -> I.cup (sigma "a") (sigma "b")
       | "b" -> fun sigma -> sigma "d" + 1
       | "c" -> fun sigma -> I.cap 0 (sigma "b")
       | _   -> fun sigma -> 1
     
let result = Test.T.to_list (Test.solve eq ["a"])
*)

