%{

open Datatypes;;
open Lexing;;
let parse_error str = let pos = Parsing.symbol_start_pos ()
		      in let file = pos.pos_fname
		      in let _ = print_string file
                      in let line = pos.pos_lnum
                      in let bol = pos.pos_bol
                      in let cnum = pos.pos_cnum
                      in let _ = List.iter print_string [
                        str;
                        ": line "; string_of_int line;
                        ", char "; string_of_int (cnum-bol);
                        "\n"]
                      in flush stdout


%}
%token LPAR RPAR
%token BEGINLIST ENDLIST
%token SEP
%token LISTSEP
%token CONS
%token IMPLIED
%token QMARK
%token ADD
%token SUB
%token MUL
%token MOD
%token IS
%token NEQ
%token EQ
%token IEQ
%token INE
%token ILT
%token IGT
%token ILE
%token IGE
%token DOT
%token NOT
%token TRUE
%token ISINT
%token ISSTRING
%token UNKNOWN
%token <string> NAME
%token <string> VAR
%token <string> STRING
%token <int> INT
%token EOF
%token HELP
%token SLASH
%token USE
%token ALL
%token DUMP
%token EXIT
%token RESET
%token OBSERVE
%token TOPDOWN
%token SOLVE
%token DIR
%token USEDIR
%token OUTDIR
%token OUTFILE

%start prog
%type <Datatypes.command list> prog

%%

prog : 
|	clauses prog1			{ Prog (List.rev $1) :: $2 }
|	prog1				{ $1 }
;

prog1 :
 	command prog			{ $1 :: $2 }
| 	command 			{ [$1] }
|	QMARK body 			{ [Question $2] }
|	TOPDOWN 			{ [Topdown] }
|	SOLVE 				{ [Solve] }
|	HELP				{ [Help] }
|	DUMP				{ [Dump] }
|	EXIT				{ exit 0 }
|	EOF				{ [Eof] }
;

command :
	USE STRING 			{ Use $2 }
|	ALL 				{ All }
|	OBSERVE NAME SLASH INT		{ Observe ($2,$4) }
|	DIR STRING   			{ Dir $2 }
|	USEDIR STRING   		{ Usedir $2 }
|	OUTDIR STRING   		{ Outdir $2 }
|	OUTFILE STRING   		{ Outfile $2 }
|	RESET   			{ Reset }
;

clauses :
	clauses clause			{ $2 :: $1 }
|	clause				{ $1 :: [] }
;

clause :
	NAME args IMPLIED body 		{ Clause ($1,$2,$4) }
|	NAME args DOT			{ Clause ($1,$2,[]) }
;

body :
	query body1			{ $1 :: $2 }
|	TRUE body1			{ $2 }
|	DOT				{ [] }
;

body1 :
	SEP query body1			{ $2 :: $3 }
|	SEP TRUE body1			{ $3 }
|	DOT				{ [] }
;

query0:
	NAME args			{ Query ($1, $2) }
;

query : 
	query0				{ $1 }
|	ISINT LPAR VAR RPAR		{ IsInt $3 }
|	ISSTRING LPAR VAR RPAR		{ IsString $3 }
|	VAR EQ UNKNOWN			{ True }
|	STRING EQ UNKNOWN		{ True }
|	INT EQ UNKNOWN			{ True }
|	UNKNOWN EQ UNKNOWN		{ True }
|	NAME EQ UNKNOWN			{ True }
|	query0 EQ UNKNOWN		{ True }
|	UNKNOWN EQ VAR			{ True }
|	UNKNOWN EQ tree0		{ True }
|	VAR EQ VAR			{ Eq (Var $1, Var $3) }
|	VAR EQ tree0			{ Eq (Var $1, Const $3) }
|	STRING EQ VAR			{ Eq (Const (String $1), Var $3) }
|	STRING EQ tree0			{ Eq (Const (String $1), Const $3) }
|	INT EQ VAR			{ Eq (Const (Int $1), Var $3) }
|	INT EQ tree0			{ Eq (Const (Int $1), Const $3) }
|	SUB INT EQ VAR			{ Eq (Const (Int (- $2)), Var $4) }
|	SUB INT EQ tree0		{ Eq (Const (Int (- $2)), Const $4) }
|	NAME EQ VAR			{ Eq (Const (Atom $1), Var $3) }
|	NAME EQ tree0			{ Eq (Const (Atom $1), Const $3) }
|	treelist EQ VAR                 { Eq (Const $1, Var $3) }
|	treelist EQ tree0		{ Eq (Const $1, Const $3) }
|	query0 EQ VAR			{ let do_arg = function
						Const t -> t
						| Var x -> TVar x
						| Unknown -> TUnknown
					  in match $1
					  with Query (r,args) -> let list = List.map do_arg args
							in Eq (Const (Cons (r,list)), Var $3)
					   | _ -> raise (Failure "internal parse error?")
					}
|	query0 EQ tree0                   { let do_arg = function
                                                Const t -> t
                                                | Var x -> TVar x
                                                | Unknown -> TUnknown
                                          in match $1
                                          with Query (r,args) -> let list = List.map do_arg args
                                                        in Eq (Const (Cons (r,list)), Const $3)
                                           | _ -> raise (Failure "internal parse error?")
                                        }
|	VAR NEQ UNKNOWN			{ raise (Failure "Unknown in disequality!") }
|	INT NEQ UNKNOWN			{ raise (Failure "Unknown in disequality!") }
|	SUB INT NEQ UNKNOWN		{ raise (Failure "Unknown in disequality!") }
|	UNKNOWN NEQ UNKNOWN		{ raise (Failure "Unknown in disequality!") }
|	NAME NEQ UNKNOWN		{ raise (Failure "Unknown in disequality!") }
|	query0 NEQ UNKNOWN		{ raise (Failure "Unknown in disequality!") }
|	UNKNOWN NEQ VAR			{ raise (Failure "Unknown in disequality!") }
|	UNKNOWN NEQ tree0		{ raise (Failure "Unknown in disequality!") }
|	VAR NEQ VAR			{ Neq (Var $1, Var $3) }
|	VAR NEQ tree0			{ Neq (Var $1, Const $3) }
|       STRING NEQ VAR                  { Neq (Const (String $1), Var $3) }
|       STRING NEQ tree0                { Neq (Const (String $1), Const $3) }
|       INT NEQ VAR                     { Neq (Const (Int $1), Var $3) }
|       INT NEQ tree0                   { Neq (Const (Int $1), Const $3) }
|       SUB INT NEQ VAR                 { Neq (Const (Int (- $2)), Var $4) }
|       SUB INT NEQ tree0               { Neq (Const (Int (- $2)), Const $4) }
|	NAME NEQ VAR			{ Neq (Const (Atom $1), Var $3) }
|	NAME NEQ tree0			{ Neq (Const (Atom $1), Const $3) }
|	treelist NEQ VAR                { Neq (Const $1, Var $3) }
|	treelist NEQ tree0		{ Neq (Const $1, Const $3) }
|	query0 NEQ VAR			{ let do_arg = function
						Const t -> t
						| Var x -> TVar x
						| Unknown -> TUnknown
					  in match $1
					  with Query (r,args) -> let list = List.map do_arg args
							in Neq (Const (Cons (r,list)), Var $3)
					   | _ -> raise (Failure "internal parse error?")
					}
|	query0 NEQ tree0		{ let do_arg = function
                                                Const t -> t
                                                | Var x -> TVar x
                                                | Unknown -> TUnknown
                                          in match $1
                                          with Query (r,args) -> let list = List.map do_arg args
                                                        in Neq (Const (Cons (r,list)), Const $3)
                                           | _ -> raise (Failure "internal parse error?")
					}
|	VAR IS exp			{ Is ($1, $3) }
|	UNKNOWN IS exp 			{ True }
|	exp comp exp			{ Comp ($2, $1, $3) }
|	NOT LPAR NAME args RPAR		{ Neg ($3, $4) }
|	NOT LPAR ISINT LPAR VAR RPAR RPAR		{ NegInt $5 }
|	NOT LPAR ISSTRING LPAR VAR RPAR RPAR		{ NegString $5 }
;

comp :
	IEQ				{ E }
|	INE				{ Ne }
|	ILE				{ Le }
|	IGE				{ Ge }
|	ILT				{ Lt }
|	IGT				{ Gt }
;

exp :
	exp MUL exp1			{ Mul ($1, $3) }
|	exp SLASH exp1			{ Div ($1, $3) }
|	exp MOD	exp1			{ Mod ($1, $3) }
|	exp1				{ $1 }
;

exp1:
	exp1 ADD exp2			{ Add ($1, $3) }
|	exp1 SUB exp2			{ Sub ($1, $3) }
|	exp2				{ $1 }
;

exp2 :
	VAR				{ EVar $1 }
|	INT				{ EInt $1 }
|	SUB exp2			{ Minus $2 }
|	LPAR exp RPAR			{ $2 }
;

args :
	LPAR VAR args1			{ Var $2 :: $3 }
|	LPAR UNKNOWN args1		{ Unknown :: $3 }
|	LPAR tree0 args1		{ Const $2 :: $3 }
|	LPAR RPAR			{ [] }
;

args1 :
	SEP VAR args1			{ Var $2 :: $3 }
|	SEP UNKNOWN args1			{ Unknown :: $3 }
|	SEP tree0 args1			{ Const $2 :: $3 }
|	RPAR				{ [] }
;

tree0 :
	NAME				{ Atom $1 }
|	STRING				{ String $1 }
|	INT				{ Int $1 }
|	SUB INT				{ Int (- $2) }
|	NAME targs			{ Cons ($1, $2) }
|	treelist			{ $1 }
;

targs:
	LPAR tree targs1                 { $2 :: $3 }
|       LPAR RPAR                       { [] }
;

tree :
	NAME				{ Atom $1 }
|	STRING				{ String $1 }
|	VAR				{ TVar $1 }
|	UNKNOWN				{ TUnknown }
|	INT				{ Int $1 }
|	NAME targs			{ Cons ($1, $2) }
|	treelist			{ $1 }
;

targs1 :
        SEP tree targs1                  { $2 :: $3 }
|       RPAR                            { [] }
;

treelist :
CONS LPAR tree SEP tree RPAR    	{ Cons ("::",[$3;$5]) }
|	BEGINLIST ENDLIST		{ Atom "[]" }
|	BEGINLIST tree ENDLIST		{ Cons ("::", [$2; Atom "[]"]) }
|	BEGINLIST tree treelist1	{ Cons ("::", [$2; $3]) }
;

treelist1:
	LISTSEP tree treelist1		{ Cons ("::", [$2; $3]) }
|	LISTSEP tree ENDLIST		{ Cons ("::", [$2; Atom "[]"]) }
;

%%
