open Commands

(* ... the main loop ... *)

let rec main ()  = 	print_string "> "; flush stdout;
		 	(execute (fun () -> exit 0) (get_input stdin);
			main ())

let print_help () = List.iter print_string [
		"\n\tAvailable Options:\n\n";
		"\t\t - :\t\tuse standard input\n";
		"\t\t<file>:\t\tuse <file>\n";
		"\t\t-u <file>:\tuse <file>\n";
		"\t\t-f <file>:\toutput to <file>\n";
		"\t\t-d <dir>:\tset use- and filedir to <dir>\n";
		"\t\t-ud <dir>:\tset usedir to <dir>\n";
		"\t\t-fd <dir>:\tset filedir to <dir>\n";
		"\t\t-o <p/k>:\tobserve <p/k>\n";
		"\t\t-s :\t\tsolve but do not switch to command line\n";
		"\t\t-t :\t\ttopdown solve but do not switch to command line\n";
		"\t\t-q :\t\tdo not solve but switch to command line\n";
		"\t\t-sq :\t\tsolve, then switch to command line\n";
		"\t\t-tq :\t\ttopdown solve, then switch to command line\n";
		"\t\t-h :\t\tprint this message\n\n"
	]
		
let command_line () = let list = Array.to_list Sys.argv
		   in let rec do_command = function
			[] -> ()
		| "-s"::rest -> ((try solve ()
                                 with Failure str -> print_string 
					("Failure: " ^ str ^ "\n")); 
				 do_command rest)
		| "--solve"::rest -> ((try solve ()
                                 with Failure str -> print_string 
					("Failure: " ^ str ^ "\n")); 
				 do_command rest)
		| "-t"::rest -> ((try topdown ()
                                 with Failure str -> print_string 
					("Failure: " ^ str ^ "\n")); 
				 do_command rest)
		| "--topdown"::rest -> ((try topdown ()
                                 with Failure str -> print_string 
					("Failure: " ^ str ^ "\n")); 
				 do_command rest)
		| "-u":: s :: rest -> (use s; do_command rest)
		| "--use":: s :: rest -> (use s; do_command rest)
		| "-d":: s :: rest -> (set_dir s; do_command rest)
		| "--dir":: s :: rest -> (set_dir s; do_command rest)
		| "-ud":: s :: rest -> (set_usedir s; do_command rest)
		| "--usedir":: s :: rest -> (set_usedir s; do_command rest)
		| "--fd":: s :: rest -> (set_filedir s; do_command rest)
		| "--filedir":: s :: rest -> (set_filedir s; do_command rest)
		| "-f":: s :: rest -> (set_outfile s; do_command rest)
		| "--file":: s :: rest -> (set_outfile s; do_command rest)
		| "-o"::s :: rest -> let i = String.index s '/'
				  in let n = String.length s
				  in let p = String.sub s 0 i
				  in let k = int_of_string (String.sub s (i+1) (n-i-1))
				  in (observe (p,k); do_command rest)
		| "--observe"::s :: rest -> let i = String.index s '/'
                                  in let n = String.length s
                                  in let p = String.sub s 0 i
                                  in let k = int_of_string (String.sub s (i+1) (n-i-1))
                                  in (observe (p,k) ; do_command rest)
		| "-" :: rest	-> (use0 (); do_command rest)
		| "-q" :: []	-> main ()
		| "--query" :: []	-> main ()
		| "-sq" :: []	-> ((try solve ()
                               with Failure str -> print_string ("Failure: " ^ str ^ "\n")); 
			       main ())
		| "-tq" :: []	-> (try topdown ()
                               with Failure str -> print_string ("Failure: " ^ str ^ "\n"); 
			       main ())
		| "-h" :: []	-> print_help ()
		| "--help" :: []	-> print_help ()
		| s :: rest -> (use s; do_command rest)

	in match list
	with _::[] -> main ()
	| _::list -> do_command list
	| [] -> print_string "\nError: empty command line?\n\n"


 
let _ =  command_line ()
			

