open Items
open Table
open Symbol
open Datatypes
open Datalog

(*
   type tree = Int of int | String of int | TVar of int | Cons of (int * tree list) | Atom of int
*)


module TreeItem = struct
   type item = (int,int) tree
   
   let rec to_int = function
	   Int int -> 7 * int
	|  TVar x -> 13 * x
	|  TUnknown -> 101
	|  String str -> 5 * str
	|  Atom x -> 17 * x
	|  Cons (a,list) -> let rec doit acc = function
		   [] -> acc
		| t::tt -> doit (to_int t + 31 * acc) tt
		in doit (67 * a) list

   let to_string tree = "to do :-)"
end

module TreeToken = struct
        type item = Tun | Tint of int | Tstr of int | Tvar of int | Tato of int | Tcon of int
        let to_int = function
		Tun	-> 23
	|	Tint int -> 11 * int
	|	Tstr str -> 13 *  str
	|	Tvar var -> 17 * var
	| 	Tato ato -> 19 * ato
	| 	Tcon con -> 19 * con


	let to_string = function
		Tun	 -> "_"
        |       Tint int -> "Int " ^ string_of_int int
        |       Tstr str -> "String " ^ string_of_int str
        |       Tvar var -> "Var " ^ string_of_int var
        |       Tato ato -> "Atom " ^ string_of_int ato
        |       Tcon con -> "Cons " ^ string_of_int con
end


module Tree = struct
  
	module S = SymbolTable (PairItem (TreeToken) (ListItem (IntItem)))
	module TT = TreeToken

let cons t list = S.index_of (t,list)

let destr int = S.symbol_of int

(* returns an index for a given pattern t with variable bindings from eta *)

let rec construct t eta = match t
	with	Int x -> cons (TT.Tint x) []
	|	TUnknown -> cons TT.Tun []
	|	TVar x -> (match eta x
			with None -> raise (Failure "Floundering construction of term?")
			| Some i -> i
			)
	|	String str -> cons (TT.Tstr str) []
	|	Atom a -> cons (TT.Tato a) []
	|	Cons (a,succ) -> let succ = List.map (fun t -> construct t eta) succ
				   in  cons (TT.Tcon a) succ

let get_int t = construct t (fun x -> raise (Failure "Ground term expected!"))

(* infers Bindings for a given pattern t and an index denoting a tree *)

let rec destruct tree i = let eta = new_env ()
		in let rec do_tree tree i = let (t,l) = destr i
			in match (tree,t)
		with	(Int x, TT.Tint y) -> if x = y then true
						else false
		|	(TUnknown, _)	-> true
		|	(String s1, TT.Tstr s2) -> if s1 = s2 then true
						     else false
		|	(Atom a1, TT.Tato a2) -> if a1 = a2 then true
						     else false
		|	(TVar x, _) -> (match Env.sub eta x
					with Some j -> if i=j then true
						       else false
					|    None -> Env.add eta x i; 
						     true
					)
		|	(Cons (a,list), TT.Tcon b) -> let rec do_lists (list,l) = match (list,l) 
						with	([],[])	-> true
						| (a::aa,i::ii) -> if do_tree a i then do_lists (aa,ii) 
								   else false
						| _	-> false
						
					in if a = b then do_lists (list,l) 
					   else false
		|	_ -> false
		
	in if do_tree tree i then Some (Env.to_list eta)
	   else None

(* recovers a tree from the internal representation *)

let rec recover i = let (t,l) = destr i
        in match t
                with    TT.Tint y -> Int y 
		|	TT.Tun	  -> TUnknown
                |       TT.Tstr s -> String s
                |       TT.Tato a -> Atom a
                |       TT.Tvar x -> TVar x
                |       TT.Tcon b -> Cons (b, List.map recover l)

(*
let t1 = Cons (1, [TVar 5; Atom 1; Int 12345])
let t2 = String 3
let i2 = get_int t2
let eta = function 5 -> Some i2
		| _ -> None 
let t = recover (construct t1 eta)
*)
end
