open Items

open Table
open Pool


module Graph (X: Item) = struct

	module T = Table(X)
	module S = Pool(X)

let edges : X.item list T.table = T.make_table 128
let nodes : S.set = S.make_set 128

let insert_node x = S.insert nodes x
let insert_edge (x,y) = S.insert nodes x;
			S.insert nodes y;
			match T.sub edges x 
		with None   -> T.add edges x [y]
		| Some list -> T.update edges x (y::list)

let get_successors x = match T.sub edges x
		with None -> []
		| Some list -> list

let get_nodes () = S.to_list nodes

let get_finite inflist = let nodes = get_nodes ()

		 in let called = S.make_set 128
	         in let finite : bool T.table = T.make_table 128
		 in let _ = List.iter (fun x -> T.update finite x false) inflist

	         in let rec dfs x = match T.sub finite x
					with Some _ -> ()
					| None -> 
				    if S.has_item called x then 
					T.update finite x false
				    else let _ = S.insert called x
				      in let succs = get_successors x
				      in let rec do_succ = function
					     		[] -> true
                                                | y::ys -> let _ = dfs y
                                                        in match T.sub finite y
                                                           with Some true -> do_succ ys
                                                             |  Some false -> false
                                                             |  None -> raise (Failure "strange loop?")

				      in (T.update finite x (do_succ succs)
					(* ;
					  S.rem_item called x
					*)
					)
	
	         in let _ = List.iter dfs nodes
	         in let rec doit = function
		    	[] -> []
		     | (x,b) :: rest -> if b then x :: doit rest
				   else doit rest
	         in doit (T.to_list finite)
	
let reset () = T.reset_table edges 128;
	       S.reset nodes 128

end

	module G = Graph (IntItem)
	      
	let _ = List.iter G.insert_edge [(1,2); (2,3); (3,2); (3,4); (3,5); (4,5)]
(*
	let _ = List.iter G.insert_edge [(1,2); (2,3)]
*)

