module type Item =
sig
      type item
      val to_string	: item -> string
      val to_int	: item -> int
end

module UnitItem = struct
   type item = unit
   let to_string () = "()"
   let to_int () = 42
end

module BoolItem = struct
   type item = bool
   let to_string = function
	  true -> "true"
	| false -> "false"
   let to_int = function
	  true -> 17
	| false -> 91
end

module IntItem = struct
   type item = int
   let to_string = string_of_int 
   let to_int x = x
end

module StringItem = struct
   type item = string
   let to_string x = x
   let to_int x = let n = String.length x
	       in let result = ref 0
	       in let _ = for i=0 to n-1 do
			result := int_of_char x.[i] + 31 * (!result)
		  	done
	       in (!result)
end

module OptionItem (X: Item) =
struct
   type item = X.item option
   let to_string = function
	  None -> "None"
	| Some x -> "Some " ^ X.to_string x
   let to_int = function
	  None -> 17
	| Some x -> 5 * X.to_int x
end

module Sum (X: Item) (Y: Item) =
struct
   type item = First of X.item | Second of Y.item 

   let put_fst x = First x
   let put_snd y = Second y

   let get_fst = function
          First x -> x
        | _ -> raise (Failure "illegal argument of get_fst")
   let get_snd = function
          Second x -> x
        | _ -> raise (Failure "illegal argument of get_snd")

   let to_string = function
          First x  -> X.to_string x 
        | Second y -> Y.to_string y

   let to_int = function
          First x  -> X.to_int x + 42
        | Second y -> Y.to_int y
end

module PairItem (X:Item) (Y:Item) = struct
  type item = X.item * Y.item
  let to_string (x,y) = "(" ^ X.to_string x ^", "^ Y.to_string y ^")"
  let to_int (x,y) = 13 * X.to_int x + Y.to_int y
end

module ListItem (X:Item) = struct
  type item = X.item list
  let rec to_int = function
	  []    -> 19
	| x::xs -> X.to_int x + 31 * to_int xs
  let to_string list = let rec doit = function
			[] -> ["]"]
		| x :: xs  -> ("; "^X.to_string x) :: doit xs 
	in match list with 
		[] -> "[]"
	    | x::xs -> String.concat "" (("[" ^ X.to_string x) :: doit xs)
end

module ArrayItem (X: Item) = struct
   module L = ListItem(X)

   type item = X.item array

let to_int array = let list = Array.to_list array
                   in L.to_int list

let to_string array = let list = Array.to_list array
                   in "|" ^ L.to_string list ^ "|"
end

module PairInt = PairItem (IntItem) (IntItem)

