(*****************************************)
(* Librairie de gestion des debordements *)
(* Sebastien Ailleret                    *)
(*****************************************)

(* remplacement des operations standards par des operations    *)
(* qui levent une exception en cas de debordement              *)

(* Pour l'utiliser, il suffit de commencer votre programme par *)
(* open Overflow                                               *)

(* en fin de fichier se trouvent :                             *)
(*  - une version approximative pour la multiplication         *)
(*      (suggre par Pierre Weis)                             *)

exception Overflow of string

(* cas de l'addition : Merci a Xavier Leroy pour l'algo *)
let ( + ) a  b =
  if (a lxor b) lor (a lxor (lnot (a+b))) < 0  (* no kidding! *)
  then a + b
  else raise (Overflow (Printf.sprintf "%d + %d" a b))

(* cas de la soustraction *)
let ( - ) a b =
  if b = min_int
  then
    if a >= 0
    then raise (Overflow (Printf.sprintf "%d - %d" a b))
    else a - b
  else
    let oppb = - b in
    a + oppb

(* cas de la multiplication    *)
(* inspiree par Michel Quercia *)
let ( * ) a b =
  if a = 0 then 0
  else
    let x = a * b in
    if b = x/a
    then x
    else raise (Overflow (Printf.sprintf "%d * %d" a b))


(****************************************************************)
(* Autres versions possibles                                    *)
(****************************************************************)

(*******************************************************)
(* autre version qui utilise une sorte de log discret  *)
(* inspiree par Pierre Weis                            *)
(* Version approximative qui refuse trop de choses     *)
(*
let length_of_int = 31 (* mettre 32 donne un resultat trop relax *)

let num_bits_int n =
  let rec num_bits n =
    if n = 0
    then 0
    else succ (num_bits (n lsr 1)) in
  num_bits (abs n)

let ( * ) n m =
  if num_bits_int n + num_bits_int m < length_of_int
  then n * m
  else raise (Overflow (Printf.sprintf "%d * %d" a b))
*)
