open Items
open Table

module Pool (X:Item) = struct
   module T = Table (X)
   type set = unit T.table
   let make_set n = T.make_table n 
   let new_set () = T.new_table ()
   let reset set = T.reset_table (set : set) 
   let has_item set x = match T.sub set x with Some () -> true 
					|    None -> false
   let rem_item set x = T.rem_item set x
   let add_item set x = T.add set x ()
   let insert set x = T.update set x ()

   let from_list list = T.from_list (
			List.map (fun x -> (x,())) list)
   let to_list set = List.map (fun (x,()) -> x) (
		     T.to_list set)
   let copy_set set = T.copy_table set

   let cup s1 s2 = let s = copy_set s2
		in let l1 = to_list s1
		in let _ = List.iter (insert s) l1
		in s

   let cap s1 s2 = let l1 = to_list s1
		in let rec doit acc = function
			 [] -> acc
		   | x::xs -> if has_item s2 x then doit (x::acc) xs
			      else doit acc xs
		in from_list (doit [] l1)

   let is_empty s = match to_list s
		with [] -> true
		| _	-> false

   module L = ListItem(X)

   let to_int set = L.to_int (
		      to_list set)

   let to_string set =
	String.concat "" (
	List.map (fun x -> X.to_string x ^" ") (
	to_list set))
end
