open Items
open Table

exception Symbol_not_found of int

module SymbolTable (X: Item) = struct

	module I = Table (X)
	module O = Table (IntItem)

let index_table  : int I.table    = I.make_table 128
let symbol_table : X.item O.table = O.make_table 128
let count = ref 0

let new_index () = let n = !count
		   in count := n+1;
		      n

let index_of x = match I.sub index_table x
		 with Some index -> index
		 |    None -> 	   let n = new_index ()
				in I.add index_table x n;
				   O.add symbol_table n x;
				   n

let symbol_of i = match O.sub symbol_table i
		   with Some x	-> x
		   |	None	-> raise (Symbol_not_found i)
			
let has_index x = match I.sub index_table x
                 with Some _ -> true
		  | None     -> false

let reset () = (I.reset_table index_table 128;
		O.reset_table symbol_table 128;
		count := 0
		)

end

