open Items

module Table (X :Item) = struct 

      type  'b buckets = (X.item * 'b) list

      type  'b table = {	mutable hash: X.item -> int; 
				mutable buckets: 'b buckets array; 
				mutable length: int
			}

(* Buckets ... *)

      let sub_bucket bucket item = 
	  let rec doit = function
		[] -> None
	      | (key, value)::rest -> if key=item then Some value
				      else doit rest
	  in doit bucket
		
      let rem_bucket bucket item =
	  let rec doit = function
		[] -> []
	      |  (key,v)::rest -> if key=item then rest
				  else (key,v)::doit rest
	  in doit bucket
	
      let update_bucket bucket item value =
	  let rec doit = function
		[] -> (true, [(item,value)])
               |  ((a,b)::rest) -> if a=item then 
					  (false, (a,value) :: rest)
                                       else let (flag, list) = doit rest
					    in (flag, (a,b)::list)
         in doit bucket
       
      let modify_bucket f bucket = 
          let g (item,value) = (item, f value)
          in List.map g bucket
 
      let bucket_to_string value_to_string list
            = let f (item , value)
                      = "(" ^ 
                        X.to_string item ^
                        " -> " ^
                        value_to_string value ^
                        ")"
           in let doit s cca
                      = s ^ "  " ^ cca
           in List.fold_right doit (List.map f list) "\n"
					  
      let bucket_to_int value_to_int x = 
		let rec doit = function
			[] -> 11
		    | (x,d)::xs -> X.to_int x + 5 * value_to_int d + 31 * doit xs
		in doit x 
					  
(* Tables ... *)

      let hash_int n = fun x -> let y = X.to_int x mod n
                                in if y < 0 then y+n else y
      let make_table n = if n <= 0 then 
		raise (Failure ("negative table size " ^
				string_of_int n ^"??"))
	else
		{
	  hash    = hash_int n;
	  buckets = Array.create n [];
	  length   = 0
	}
      
      let new_table () = make_table 2

      let reset_table table n = 
		( table.buckets <-  Array.create n []
		; table.hash    <- hash_int n
		; table.length	  <- 0
		)

      let sub table item = 
	  let i = table.hash item
	  in sub_bucket table.buckets.(i) item

      let rem_item table item = 
	     let i = table.hash item
	  in let bucket = rem_bucket table.buckets.(i) item
	  in     table.buckets.(i) <- bucket

      let has_value table item = match sub table item
	  with Some _ -> true
	   | None   -> false

      let double_table table = 
	     let n = Array.length table.buckets
          in let new_n = 2*n
	  in let _ = if new_n < 0 then raise (Failure ("negative table length " ^
							string_of_int new_n))
		     else ()
							
	  in let _ = table.hash  <- hash_int new_n
	  in let new_buckets = Array.create new_n []

	  in let _ = 
	     for i = 0 to Array.length table.buckets - 1 do
	         let bucket = table.buckets.(i)
	      in let rec doit = function
			[] -> ()
	       |  (index,value)::xs -> 
			let j = table.hash index
		     in let list = new_buckets.(j)
		     in let _ = new_buckets.(j) <- (index,value)::list
		     in doit xs
	      in doit bucket
             done
	  in table.buckets <- new_buckets

      let add table item value = 
	  let n = Array.length table.buckets
(*
       in let _ = if (table.length >= (n * 2) / 3) then double_table table
                  else ()
*)
       in let _ = if table.length >= n then double_table table
                  else ()
       in let i = table.hash item
       in table.length <- table.length +1 ;
          table.buckets.(i) <- (item,value) :: table.buckets.(i)

      let update table item value = 
	  let n = Array.length table.buckets
       in let _ = if (table.length >= (n * 2) / 3) then double_table table
		  else ()
       in let i = table.hash item
       in let bucket = table.buckets.(i)
       in let (flag, bucket) = update_bucket bucket item value
       in let _ = table.buckets.(i) <- bucket
       in if flag then table.length <- table.length + 1
	     else ()

      let from_list list = 
	  let table = make_table (2 * List.length list +3)
       in let rec doit = function
              [] -> table
	   |  (item,value) :: rest -> (add table item value; doit rest)
       in  doit list

      let to_list table = List.concat (
			  Array.to_list (table.buckets))

     let copy_table table  = 
             let buckets = Array.init (Array.length table.buckets)
                        (fun i -> table.buckets.(i))
          in let hash = table.hash
          in let length = table.length
          in {buckets = buckets;hash = hash;length = length}

     let modify_table f table  = 
		for i = 0 to Array.length table.buckets do
          		table.buckets.(i) <- modify_bucket f table.buckets.(i)
		done
 
      let to_string value_to_string table = 
		bucket_to_string value_to_string (to_list table)
    
      let to_int value_to_int table =
		bucket_to_int value_to_int (to_list table)
   end
