open RightHandSide
module M = Mtt

exception Non_deterministic_mtt
exception Partially_defined_mtt
exception Not_top_down

let rec rhs_to_string mtt = function
   Empty_rhs -> "" |
   Parameter i -> "y_"^(string_of_int i) |
   Symbol (s,k,succ) ->
      M.delta_to_string mtt (M.delta_of_ord mtt s)^
      "("^(rhs_list mtt succ)^")" |
   Call (q,s,k,succ) ->
      M.state_to_string mtt (M.state_of_ord mtt q)^
      "(x_"^(string_of_int s)^
      (if k>0 then ","^(rhs_list mtt succ) else "")^")"
and rhs_list mtt = function
   [] -> "" |
   r::[] -> rhs_to_string mtt r |
   r::rs -> (rhs_to_string mtt r)^","^(rhs_list mtt rs)

let action_to_string mtt = function
   M.Undefined -> "undef" |
   M.Det ac -> rhs_to_string mtt ac |
   M.Non_det acs -> "["^(rhs_list mtt acs)^"]"



let num_of_states mtt =
   let n = ref 0 in
      M.iter_states (fun a b -> n := !n + 1) mtt; !n



let mkrules mtt top comp =
   let rec repl i rhs_list = function
      Empty_rhs -> Empty_rhs |
      Parameter j -> Parameter j |
      Symbol (s,k,rhs) ->
         let sym' = M.delta_of_ord top s in
         let _ = M.add_delta comp sym' in
         let sym = M.ord_of_delta comp sym' in
         let rhs' = List.map (repl i rhs_list) rhs in
         let tmp = Symbol (sym,k,rhs') in
         tmp |
      Call (q,x,k,rhs) ->
         if k > 0 then
            raise Not_top_down
         else
            let r = List.nth rhs_list (x-1) in
            let tmp = tr q r in
(*             let _ = Printf.printf "\t\t%s/tr(%i)(%s)\n==\t\t\t%s\n" *)
(*             (rhs_to_string top (Call(q,x,k,rhs))) q (rhs_to_string mtt r) *)
(*             (rhs_to_string comp tmp) in *)
            tmp
   and tr i = function
      Empty_rhs -> Empty_rhs |
      Parameter j -> 
         let new_j = (num_of_states top) * (j-1) + i + 1 in
         Parameter new_j |
      Symbol (a,k,rhs_list) ->
         let sym' = Mtt.ord_of_sigma top (Mtt.delta_of_ord mtt a) in
         let s = Mtt.get_action top i sym' in
         let s' = 
            match s with
               Mtt.Undefined -> raise Partially_defined_mtt |
               Mtt.Det rhs' -> repl i rhs_list rhs'|
               Mtt.Non_det rhs' -> raise Non_deterministic_mtt in
(*          let _ = Printf.printf "\t%s[../..], %i = %s\n" (action_to_string top s) *)
(*          i (rhs_to_string comp s')  *)
(*          in *)
            s' |
      Call (q,x,k,rhs) ->
         let t_state = Mtt.state_of_ord top i in (* to do *)
         let m_state = Mtt.state_of_ord mtt q in
         let _ = Mtt.add_state comp (m_state,t_state) in
         let q' = Mtt.ord_of_state comp (m_state,t_state) in
         let rhs' =
            let rh = ref [] in
               List.iter 
                  (fun x -> 
                     rh := !rh @ (List.rev (mk_rhs x (num_of_states top)))
                  ) rhs;
               !rh 
            in
               Call (q',x,List.length rhs',rhs')
   and mk_rhs rhs = function
      0 -> [] |
      i -> (tr (i-1) rhs) :: (mk_rhs rhs (i-1)) in
   let tr_ac i = function
      Mtt.Undefined -> raise Partially_defined_mtt |
      Mtt.Det rhs -> tr i rhs | 
      Mtt.Non_det rhs_list -> raise Non_deterministic_mtt 
   in
   Mtt.iter
      (fun (m_state,m_symbol,m_action) ->
         Mtt.add_sigma comp (Mtt.sigma_of_ord mtt m_symbol);
         Mtt.iter_states
            (fun t_state i ->
               let rhs = tr_ac i m_action in
(*                let _ = Printf.printf "%s, %i --> %s\n" (action_to_string mtt *)
(*                m_action) i (rhs_to_string comp rhs) in *)
               let a = Mtt.state_of_ord mtt m_state in
               let _ = Mtt.add_state comp (a,t_state) in
               let state = Mtt.ord_of_state comp (a,t_state) in
               let sym = M.ord_of_sigma comp (M.sigma_of_ord mtt m_symbol)
               in
                  Mtt.add_rule comp state sym rhs
            ) top
      ) mtt

let top_o_mtt top mtt =
(*    let _ = Printf.printf "top o mtt\n" in *)
(*    let _ = Printf.printf "states of mtt: "; M.print_states mtt in *)
(*    let _ = Printf.printf "states of top: "; M.print_states top in *)
   let new_eq (q1,p1) (q2,p2) =
      let eq1 = Mtt.state_eq mtt in
      let eq2 = Mtt.state_eq top in
         (eq1 q1 q2) && (eq2 p1 p2) in
   let new_to_string (q,p) =
      let to_string1 = Mtt.state_to_string mtt in
      let to_string2 = Mtt.state_to_string top in
         "<" ^ (to_string1 q) ^ "," ^ (to_string2 p) ^ ">" in
   let new_from_string s =
      let from_string1 = Mtt.state_from_string mtt in
      let from_string2 = Mtt.state_from_string top in
      let i = String.index s ':' in
      let s1 = String.sub s 0 (i-1) in
      let s2 = String.sub s (i+1) (String.length s - i) in
         (from_string1 s1, from_string2 s2) in
   let comp = 
      Mtt.empty 
         (new_eq, new_to_string, new_from_string) 
         (Mtt.sigma_eq mtt, Mtt.sigma_to_string mtt, Mtt.sigma_from_string mtt)
         (Mtt.delta_eq top, Mtt.delta_to_string top, Mtt.delta_from_string top) in
   let m_init = Mtt.state_of_ord mtt (List.hd (Mtt.get_init mtt)) in
   let t_init = Mtt.state_of_ord top (List.hd (Mtt.get_init top)) in
   let init = (m_init,t_init)
   in
      Mtt.add_state comp init;
      Mtt.set_init comp (Mtt.ord_of_state comp init);
(*       Printf.printf "***** making rules...\n"; *)
      mkrules mtt top comp;
(*       Printf.printf "***** finished making rules...\n"; *)
      comp




exception Mtt_is_not_non_copying     
exception Mtt_is_not_weakly_single_use
module C = Hashtbl

let rec annotate i pos = function
   Empty_rhs -> (Empty_rhs_a (pos,false), false) |
   Parameter j -> (Parameter_a (pos,i==j,j), i==j) |
   Symbol (s,k,rhs_list) ->
      let (rhs_list', contains) = annotate' i 1 false [] rhs_list in
         (Symbol_a (pos,contains,s,k,List.rev rhs_list'), contains) |
   Call (q,x,k,rhs_list) ->
      let (rhs_list', contains) = annotate' i 1 false [] rhs_list in
         (Call_a (pos,contains,q,x,k,List.rev rhs_list'), contains)
and annotate' i pos contains result = function
   [] -> (result, contains) |
   rhs::l ->
      let (rhs', contains') = annotate i pos rhs in
         annotate' i (pos+1) (contains || contains') (rhs' :: result) l

let rec copy = function
   Empty_rhs_a (p,b) -> Empty_rhs' |
   Parameter_a (p,b,j) -> Parameter' j |
   Symbol_a (p,b,s,k,rhs_list) -> Symbol' (s,k, List.map copy rhs_list) |
   Call_a (p,b,q,x,k,rhs_list) -> Call' (q,x,k, List.map copy rhs_list)
and rewrite succ = function
   Empty_rhs_a (p,b) -> Empty_rhs' |
   Parameter_a (p,b,j) ->
      if b then Parameter_up (p,j,succ)
      else Parameter' j |
   Symbol_a (p,b,s,k,rhs_list) ->
      if b then
         let (rhs_list',father) = get [] (Empty_rhs_a (0,false)) rhs_list in
         let sons = List.map copy rhs_list' in
            rewrite (Symbol_up (p,s,k+1,succ :: sons)) father
      else
         Symbol' (s,k, List.map copy rhs_list) |
   Call_a (p,b,q,x,k,rhs_list) ->
      if b then
         let (rhs_list',father) = get [] (Empty_rhs_a (0,false)) rhs_list in
         let sons = List.map copy rhs_list' in
            rewrite (Call_up (p,q,x,k+1,succ :: sons)) father
      else
         Call' (q,x,k, List.map copy rhs_list)
and get l father = function
   [] -> (l,father) |
   x::xs ->
      match x with
         Parameter_a (_,true,_) -> get l x xs |
         Symbol_a (_,true,_,_,_) -> get l x xs |
         Call_a (_,true,_,_,_,_) -> get l x xs |
         _ -> get (x::l) father xs




let mk_rules2 m1 m2 comp =
   let rec repl i rhs_list = function
      Empty_rhs -> Empty_rhs |
      Parameter j -> Parameter j |
      Symbol (s,k,rhs) ->
         let sym' = M.delta_of_ord m2 s in
         let _ = M.add_delta comp sym' in
         let sym = M.ord_of_delta comp sym' in
         let rhs' = List.map (repl i rhs_list) rhs in
         let tmp = Symbol (sym,k,rhs') in
         tmp |
      Call (q,x,k,rhs) ->
         if k > 0 then
            raise Not_top_down
         else
            let r = List.nth rhs_list (x-1) in
            let tmp = tr q r in
(*             let _ = Printf.printf "\t\t%s/tr(%i)(%s)\n==\t\t\t%s\n" *)
(*             (rhs_to_string m2 (Call(q,x,k,rhs))) q (rhs_to_string m1 r) *)
(*             (rhs_to_string comp tmp) in *)
            tmp
   and tr i = function
      Empty_rhs -> Empty_rhs |
      Parameter j -> 
         let new_j = (num_of_states m2) * (j-1) + i + 1 in
         Parameter new_j |
      Symbol (a,k,rhs_list) ->
         let sym' = Mtt.ord_of_sigma m2 (Mtt.delta_of_ord m1 a) in
         let s = Mtt.get_action m2 i sym' in
         let s' = 
            match s with
               Mtt.Undefined -> raise Partially_defined_mtt |
               Mtt.Det rhs' -> repl i rhs_list rhs'|
               Mtt.Non_det rhs' -> raise Non_deterministic_mtt in
(*          let _ = Printf.printf "\t%s[../..], %i = %s\n" (action_to_string m2 s) *)
(*          i (rhs_to_string comp s')  *)
(*          in *)
            s' |
      Call (q,x,k,rhs) ->
         let m2_state = Mtt.state_of_ord m2 i in (* to do *)
         let m1_state = Mtt.state_of_ord m1 q in
         let _ = Mtt.add_state comp (m1_state,m2_state) in
         let q' = Mtt.ord_of_state comp (m1_state,m2_state) in
         let rhs' =
            let rh = ref [] in
               List.iter 
                  (fun x -> 
                     rh := !rh @ (List.rev (mk_rhs x (num_of_states m2)))
                  ) rhs;
               !rh 
            in
               Call (q',x,List.length rhs',rhs')
   and mk_rhs rhs = function
      0 -> [] |
      i -> (tr (i-1) rhs) :: (mk_rhs rhs (i-1)) in
   let tr_ac i = function
      Mtt.Undefined -> raise Partially_defined_mtt |
      Mtt.Det rhs -> tr i rhs | 
      Mtt.Non_det rhs_list -> raise Non_deterministic_mtt 
   in
   Mtt.iter
      (fun (m1_state,m1_symbol,m1_action) ->
         Mtt.add_sigma comp (Mtt.sigma_of_ord m1 m1_symbol);
         Mtt.iter_states
            (fun m2_state i ->
               let rhs = tr_ac i m1_action in
               let _ = Printf.printf "%s, %i --> %s\n" (action_to_string m1
               m1_action) i (rhs_to_string comp rhs) in
               let a = Mtt.state_of_ord m1 m1_state in
               let _ = Mtt.add_state comp (a,m2_state) in
               let state = Mtt.ord_of_state comp (a,m2_state) in
               let sym = M.ord_of_sigma comp (M.sigma_of_ord m1 m1_symbol)
               in
                  Mtt.add_rule comp state sym rhs
            ) m2
      ) m1



let rec contains i = function
   Parameter j -> i==j |
   Symbol (s,k,rhs) ->
      List.fold_right (fun x y -> contains i x || y) rhs false |
   Call (q,x,k,rhs) ->
      List.fold_right (fun x y -> contains i x || y) rhs false |
   _ -> false

let get_rhs = function
   Mtt.Undefined -> raise Partially_defined_mtt |
   Mtt.Non_det _ -> raise Non_deterministic_mtt |
   Mtt.Det rhs -> rhs

let num_of_pars mtt =
   let r = ref 0 in
   (Mtt.iter_states
      (fun q i -> r := !r + (Mtt.num_of_parameters mtt i)) mtt; !r)

let mk_rules mtt1 mtt2 comp =
   let c = C.create 101 in
   let mk_state k q1 j q2 =
      let s = (string_of_int k) ^"_"^ (Mtt.state_to_string mtt1 q1) ^":"^
              (string_of_int j) ^"_"^ (Mtt.state_to_string mtt2 q2) in
      Mtt.state_from_string comp s in

   let rec nf q2 rhs_q1_s l =
      let _ = Printf.printf "nf\n" in
      let q2' = Mtt.ord_of_state mtt2 q2 in
      match rhs_q1_s with
         Symbol (d,k,rhs_list) -> 
            let d'' = (Mtt.delta_of_ord mtt1 d) in
            let _ = Printf.printf "--> %s <--\n" (Mtt.delta_to_string mtt1 d'') in
            let d' = Mtt.ord_of_sigma mtt2 d'' in
            let rhs_q2_d = get_rhs (Mtt.get_action mtt2 q2' d') in
            trans rhs_list rhs_q2_d |
         Parameter k -> Parameter (k*q2') |
         Call (q,x,k,rhs_list) ->
            let q1 = Mtt.state_of_ord mtt1 q in
            let _ = Mtt.add_state comp (q1,q2) in
            let state = Mtt.ord_of_state comp (q1,q2) in
            let r = (Mtt.num_of_parameters mtt1 q) in
            let _ = Printf.printf "number of parameters of q in M1: %i\n" r in
            let mu = num_of_states mtt2 in
            let r2 = (Mtt.num_of_parameters mtt2 q2') in
            let _ = Printf.printf "number of parameters of q in M2: %i\n" r in
            let pars = r * mu + r2 in
            let _ = Printf.printf "number of parameters: %i\n" pars in
            let _ = Mtt.set_parameters comp state pars in
            let rhs_list'' = 
               nest'' (q1,q2) x [] rhs_list (List.length rhs_list) in
            let rhs_list' = rhs_list'' @ l in
            let k' = List.length rhs_list' in
               Call (state,x,k',rhs_list') |
         _ -> Empty_rhs

   and nest'' (q1,g) x l rhs_l = function
      0 -> l |
      r ->
         let _ = Printf.printf "nest''\n" in
         let l' = ref [] in
         Mtt.iter_states
            (fun q i -> l' := (nest (q1,g) r q x rhs_l) :: (!l')) mtt2;
         nest'' (q1,g) x ((List.rev (!l')) @ l) rhs_l (r-1)

   and trans l rhs_q2_d =
      let _ = Printf.printf "trans\n" in
      match rhs_q2_d with
         Call (q,x,k,rhs_list) ->
            let rhs' = List.nth l (x-1) in
            let rhs_list' = List.map (trans l) rhs_list in
               nf (Mtt.state_of_ord mtt2 q) rhs' rhs_list' |
         Symbol (s,k,rhs_list) ->
            let s' = Mtt.delta_of_ord mtt2 s in
            let _ = Mtt.add_delta comp s' in
            let sym = Mtt.ord_of_delta comp s' in
            let rhs_list' = List.map (trans l) rhs_list in
               Symbol (sym,k,rhs_list') |
         Parameter j -> Parameter j |
         _ -> Empty_rhs

   and nest (q1,g) k q2 x l =
      let _ = Printf.printf "nest\n" in
      if C.mem c (k,q2) then
         Empty_rhs
      else
         let _ = C.add c (k,q2) true in
         let r = List.length l in
         let s = Mtt.num_of_parameters mtt2 (Mtt.ord_of_state mtt2 q2) in
         let rhs_list' = mk_nest (q1,g,k,q2) x [] l r in
         nf q2 (List.nth l (k-1)) rhs_list'

   and mk_nest (q1,g,k,q2) x l rhs_l = function
      0 -> l |
      s -> 
         let _ = Printf.printf "mk_nest\n" in
         let q_sym = mk_state k q1 s q2 in
         let _ = Mtt.add_state comp q_sym in
         let q = Mtt.ord_of_state comp q_sym in
         let _ = Printf.printf "new symbol: %s\n" (Mtt.state_to_string comp
         q_sym) in
         let rhs_list'' = nest' (q1,g,k,q2) x [] rhs_l (List.length rhs_l) in
         let empty_l =
            let len = num_of_pars mtt2 in
            let l' = ref [] in
            (for j = 1 to len do
               l' := Empty_rhs :: (!l')
            done; (!l')) in
         let rhs_list' = rhs_list'' @ empty_l in
         let k = List.length rhs_list' in
         mk_nest (q1,g,k,q2) x ((Call (q,x,k,rhs_list')) :: l) rhs_l (s-1)

   and nest' (q1,g,k,q2) x l rhs_l = function
      0 -> l |
      r -> 
         let _ = Printf.printf "nest'\n" in
         let _ = C.add c (k,q2) true in
         let l' = ref [] in
         Mtt.iter_states
            (fun q i -> l' := (nest (q1,g) r q x rhs_l) :: (!l')) mtt2;
         nest' (q1,g,k,q2) x ((List.rev (!l')) @ l) rhs_l (r-1)
   in

   Mtt.iter
      (fun (m1_state,m1_symbol,m1_action) ->
         let sym = M.sigma_of_ord mtt1 m1_symbol in
         let _ = Mtt.add_sigma comp sym in
         let symbol = M.ord_of_sigma comp sym in
         let q1 = Mtt.state_of_ord mtt1 m1_state in
         Mtt.iter_states
            (fun q2 i ->
               let _ = C.clear c in
               let _ = Mtt.add_state comp (q1,q2) in
               let state = Mtt.ord_of_state comp (q1,q2) in
               let pars = Mtt.num_of_parameters mtt2 i in
               let rhs_l =
                  let l = ref [] in
                  for j = 1 to i do
                     l := (Parameter (i-j+1)) :: (!l)
                  done; (!l) in
               let rhs = nf q2 (get_rhs m1_action) rhs_l in
               Mtt.add_rule comp state symbol rhs
            ) mtt2
      ) mtt1


let mtt_o_mtt m2 m1 =
        
   let _ =
      (Printf.printf "Sigma of mtt1:\n";
      Mtt.iter_sigma
         (fun s i -> Printf.printf "%s --> %i\n" 
             (Mtt.sigma_to_string m1 s) i) m1) in
   let _ =
      (Printf.printf "Delta of mtt1:\n";
      Mtt.iter_delta
         (fun s i -> Printf.printf "%s --> %i\n" 
             (Mtt.delta_to_string m1 s) i) m1) in
   let _ =
      (Printf.printf "Sigma of mtt2:\n";
      Mtt.iter_sigma
         (fun s i -> Printf.printf "%s --> %i\n" 
             (Mtt.sigma_to_string m2 s) i) m2) in
   let _ =
      (Printf.printf "Delta of mtt2:\n";
      Mtt.iter_delta
         (fun s i -> Printf.printf "%s --> %i\n" 
             (Mtt.delta_to_string m2 s) i) m2) in
   let _ = Printf.printf "\n" in

   let b1 = AnalyseMtt.is_non_copying m1 in
   let b2 = AnalyseMtt.is_weakly_single_use m2 in
   let new_eq (q1,p1) (q2,p2) =
      let eq1 = Mtt.state_eq m1 in
      let eq2 = Mtt.state_eq m2 in
         (eq1 q1 q2) && (eq2 p1 p2) in
   let new_to_string (q,p) =
      let to_string1 = Mtt.state_to_string m1 in
      let to_string2 = Mtt.state_to_string m2 in
         "<" ^ (to_string1 q) ^ "," ^ (to_string2 p) ^ ">" in
   let new_from_string s =
      let from_string1 = Mtt.state_from_string m1 in
      let from_string2 = Mtt.state_from_string m2 in
      let i = String.index s ':' in
      let s1 = String.sub s 0 (i) in
      let s2 = String.sub s (i+1) (String.length s - i - 1) in
         (from_string1 s1, from_string2 s2) in
   let comp = 
      Mtt.empty 
         (new_eq, new_to_string, new_from_string) 
         (Mtt.sigma_eq m1, Mtt.sigma_to_string m1, Mtt.sigma_from_string m1)
         (Mtt.delta_eq m2, Mtt.delta_to_string m2, Mtt.delta_from_string m2) in
   let m_init = Mtt.state_of_ord m1 (List.hd (Mtt.get_init m1)) in
   let t_init = Mtt.state_of_ord m2 (List.hd (Mtt.get_init m2)) in
   let init = (m_init,t_init) in
      if b1 && b2 then
         (Mtt.add_state comp init;
         Mtt.set_init comp (Mtt.ord_of_state comp init);
         mk_rules m1 m2 comp;
         comp)
      else
         if not b1 then
            raise Mtt_is_not_non_copying
         else
            raise Mtt_is_not_weakly_single_use
