module type Dictionary =
   sig
      type ('a,'b) dictionary
      val create : unit -> ('a,'b) dictionary
      val add : ('a,'b) dictionary -> 'a * 'b -> unit
      val rm : ('a,'b) dictionary -> 'a -> unit
      val rm_rev : ('a,'b) dictionary -> 'b -> unit
      val replace : ('a,'b) dictionary -> 'a -> 'b -> unit
      val replace_rev : ('a,'b) dictionary -> 'b -> 'a -> unit
      val mem : ('a,'b) dictionary -> 'a -> bool
      val mem_rev : ('a,'b) dictionary -> 'b -> bool
      val lookup : ('a,'b) dictionary -> 'a -> 'b
      val lookup_rev : ('a,'b) dictionary -> 'b -> 'a
      val iter : ('a -> 'b -> unit) -> ('a,'b) dictionary -> unit
      val iter_rev : ('b -> 'a -> unit) -> ('a,'b) dictionary -> unit
      val fold : ('a -> 'b -> 'c -> 'c) -> ('a,'b) dictionary -> 'c -> 'c
      val to_list : ('a,'b) dictionary -> 'a list
      val to_list_rev : ('a,'b) dictionary -> 'b list
      val size : ('a,'b) dictionary -> int
   end

module H = Hashtbl

exception First_entry_exists
exception Second_entry_exists
exception Both_entries_exist

type ('a,'b) dictionary = 
   {
      mutable atob : ('a,'b) H.t;
      mutable btoa : ('b,'a) H.t;
      mutable size : int
   }

let create () =
   {
      atob = H.create 101;
      btoa = H.create 101;
      size = 0
   }

let inc_size dict = dict.size <- dict.size + 1

let dec_size dict = dict.size <- dict.size - 1

let add dict (a,b) =
   let a' = H.mem dict.atob a
   and b' = H.mem dict.btoa b
   in
      match (a',b') with
         (false,false) -> 
            H.add dict.atob a b; H.add dict.btoa b a; inc_size dict |
         (true,false) -> raise First_entry_exists |
         (false,true) -> raise Second_entry_exists |
         _ -> raise Both_entries_exist

let rm dict a = 
   let b' = H.find dict.atob a
   in
      H.remove dict.atob a; H.remove dict.btoa b'; dec_size dict

let rm_rev dict b =
   let a' = H.find dict.btoa b
   in
      H.remove dict.btoa b; H.remove dict.atob a'; dec_size dict

let replace dict a b = H.replace dict.atob a b 

let replace_rev dict b a = H.replace dict.btoa b a

let mem dict a = H.mem dict.atob a

let mem_rev dict b = H.mem dict.btoa b

let lookup dict a = H.find dict.atob a

let lookup_rev dict b = H.find dict.btoa b

let iter f dict = H.iter f dict.atob

let iter_rev f dict = H.iter f dict.btoa

let fold f dict init = H.fold f dict.atob init

let to_list dict = 
   let r = ref [] in H.iter (fun a _ -> r := a :: !r) dict.atob; !r

let to_list_rev dict = 
   let r = ref [] in H.iter (fun b _ -> r := b :: !r) dict.btoa; !r

let size dict = dict.size

