open Mtt
open RightHandSide

exception Execution_error
exception Undefined_action
exception Undefined_action_error
exception Empty_nondet_choice
exception Not_enough_parameters 
exception Not_enough_successors_in_input

let rec rhs_to_string mtt = function
   Empty_rhs -> "nil" |
   Parameter i -> "y_"^(string_of_int i) |
   Symbol (s,k,succ) ->
      delta_to_string mtt (delta_of_ord mtt s)^
      "("^(rhs_list mtt succ)^")" |
   Call (q,s,k,succ) ->
      state_to_string mtt (state_of_ord mtt q)^
      "(x_"^(string_of_int s)^
      (if k>0 then ","^(rhs_list mtt succ) else "")^")"
and rhs_list mtt = function
   [] -> "" |
   r::[] -> rhs_to_string mtt r |
   r::rs -> (rhs_to_string mtt r)^","^(rhs_list mtt rs)

let rec pars_to_string n = function
   0 -> "" |
   1 -> "y_" ^ (string_of_int n) |
   i -> "y_" ^ (string_of_int (n-i+1)) ^ "," ^ (pars_to_string n (i-1))

let action_to_string mtt = function
   Undefined -> "undef" |
   Det ac -> rhs_to_string mtt ac |
   Non_det acs -> "["^(rhs_list mtt acs)^"]"



   

let rec emul_action t pars mtt = function
   Undefined -> raise Undefined_action |
   Det rhs -> emul_rhs t pars mtt rhs |
   Non_det rhs_list -> 
      if (List.length rhs_list) > 0 then
         emul_rhs t pars mtt (List.hd rhs_list)
      else
         raise Empty_nondet_choice

and emul_rhs t pars mtt = function
   Empty_rhs -> Tree_input.Empty_tree |
   Parameter i -> 

      let _ = Printf.printf "(P) current tree:\n%s" (Tree_input.to_string t (fun x ->
              (sigma_to_string mtt) x)) in

      if (List.length pars) >= i then
         List.nth pars (i-1)
      else
         raise Not_enough_parameters |
   Symbol (s,k,rhs_list) ->

      let _ = Printf.printf "(S) current tree:\n%s" (Tree_input.to_string t (fun x ->
              (sigma_to_string mtt) x)) in

      let label = delta_of_ord mtt s in
      let succ = List.map (fun r -> emul_rhs t pars mtt r) rhs_list in
      Tree_input.Node (label,succ) |
   Call (q,x,k,rhs_list) ->

      let _ = Printf.printf "(C) current rhs:%s\n" (rhs_to_string mtt
              (Call(q,x,k,rhs_list))) in
      let _ = Printf.printf "(C) current tree:\n%s" (Tree_input.to_string t (fun x ->
              (sigma_to_string mtt) x)) in

      let next_tree = 
         try
            Tree_input.nth (x-1) t
         with
            Failure e -> raise Not_enough_successors_in_input in
      let next_sym = ord_of_sigma mtt (Tree_input.label next_tree) in
      let next_act = get_action mtt q next_sym in
      let next_pars = List.map (fun r -> emul_rhs t pars mtt r) rhs_list in
         emul_action next_tree next_pars mtt next_act




let run t pars mtt =

   let _ = Mtt.iter (fun (q,s,a) ->
           let state = Mtt.state_to_string mtt (Mtt.state_of_ord mtt q) in
           let sym = Mtt.sigma_to_string mtt (Mtt.sigma_of_ord mtt s) in
           let n = Mtt.num_of_parameters mtt q in
           let pars = pars_to_string n n in
           let act = action_to_string mtt a in
           if n = 0 then
              Printf.printf "%s(%s) -> %s\n" state sym act
           else
              Printf.printf "%s(%s,%s) -> %s\n" state sym pars act) mtt in


   let start = List.hd (get_init mtt) in
   let sym = ord_of_sigma mtt (Tree_input.label t) in
   let action = get_action mtt start sym in
   let _ = action_to_string mtt action in
   let rhs =      
      match action with
         Undefined -> raise Undefined_action_error |
         Det ac -> ac |
         Non_det [] -> raise Execution_error |
         Non_det (ac::acs) -> ac
   in
      emul_action t pars mtt action
