open Mtt
open RightHandSide
module A = Bottom_up_automaton

let get_rules mtt (fsta,enc) symbol =
   let symbol' = Mtt.sigma_of_ord mtt symbol in
   let fsta_symbol = enc symbol' in
   let rules = ref [] in
   A.iter (fun (symbol,qs,qs') ->
      if symbol = fsta_symbol then
         rules := (symbol,qs,qs') :: (!rules)
   ) fsta; (!rules)

let get_rules' mtt (fsta,enc) symbol state =
   let symbol' = Mtt.sigma_of_ord mtt symbol in
   let fsta_symbol = enc symbol' in
   let rules = ref [] in
   A.iter (fun (symbol,qs,qs') ->
      if (symbol = fsta_symbol) && (List.mem state qs') then
         rules := (symbol,qs,qs') :: (!rules)
   ) fsta; (!rules)
   

let rec trans mtt mtt' (fsta,enc) args tab = function
   | Empty_rhs -> Empty_rhs
   | Parameter i -> Parameter i 
   | Symbol (s,k,rhs_list) -> 
       let symb = Mtt.delta_of_ord mtt s in
       let _ = Mtt.add_delta mtt' symb in
       let new_symb = Mtt.ord_of_delta mtt' symb in
       Symbol (new_symb,k,List.map (trans mtt mtt' (fsta,enc) args tab) rhs_list)
   | Call (q,x,k,rhs_list) ->
       let new_q = (List.nth args (x-1),Mtt.state_of_ord mtt q) in
       let _ = add_rules mtt mtt' (fsta,enc) q [List.nth args (x-1)] tab in 
       let _ = Mtt.add_state mtt' new_q in
       let q' = Mtt.ord_of_state mtt' new_q in
       let _ = Mtt.set_parameters mtt' q' (Mtt.num_of_parameters mtt q) in
       Call (q',x,k,List.map (trans mtt mtt' (fsta,enc) args tab) rhs_list)

and add_rules mtt mtt' (fsta,enc) q states tab =
   let dom = 
      match (Mtt.get_domain mtt q) with
         | OneOf l -> l
         | _ -> [] 
   in
   List.iter (fun symbol ->
      let state_sym = Mtt.state_of_ord mtt q in
      let s = Mtt.sigma_of_ord mtt symbol in
      let _ = Mtt.add_sigma mtt' s (Mtt.arity mtt symbol) in
      let new_symb = Mtt.ord_of_sigma mtt' s in
      if not (Simple_table.mem tab (q,states,symbol)) then
         let _ = Simple_table.add tab ((q,states,symbol),true) in
         List.iter (fun state ->
            let rules = get_rules' mtt (fsta,enc) symbol state in
            List.iter (fun (_,qs,_) ->
               let new_state = (state,state_sym) in
               let _ = Mtt.add_state mtt' new_state in
               let state' = (Mtt.ord_of_state mtt') new_state in
               let _ = Mtt.set_parameters mtt' state' (Mtt.num_of_parameters mtt q) in
               match (Mtt.get_action mtt q symbol) with
                  | Undefined -> ()
                  | Det rhs ->
                        let rhs' = trans mtt mtt' (fsta,enc) qs tab rhs in
                        Mtt.add_rule mtt' state' new_symb rhs'
                  | Non_det rhs_list ->
                        List.iter (fun rhs ->
                           let rhs' = trans mtt mtt' (fsta,enc) qs tab rhs in
                           Mtt.add_rule mtt' state' new_symb rhs'
                        ) rhs_list
            ) rules
         ) states
   ) dom
   
let inter mtt (fsta,enc) =
   let eq (p,q) (p',q') = p=p' && (Mtt.state_eq mtt) q q' in
   let to_str (p,q) = "<"^(string_of_int p)^",q"^">" in
   let from_str s = 
      let i = String.index s ',' in
      let s1 = String.sub s 0 i in
      let s2 = String.sub s (i+1) (String.length s -i - 1) in
      (int_of_string s1,(Mtt.state_from_string mtt) s2) in
   let mtt' = 
      Mtt.empty 
         (eq,to_str,from_str)
         (Mtt.sigma_eq mtt,Mtt.sigma_to_string mtt,Mtt.sigma_from_string mtt) 
         (Mtt.delta_eq mtt,Mtt.delta_to_string mtt,Mtt.delta_from_string mtt) in
   let finals = A.final_states fsta in
   let init = List.hd (Mtt.get_initial_state mtt) in
   let tab = Simple_table.create () in
   let _ = add_rules mtt mtt' (fsta,enc) init finals tab in

(*    Mtt.iter (fun (state,symbol,action) -> *)
(*       let stat = Mtt.state_of_ord mtt state in *)
(*       let rules = get_rules mtt (fsta,enc) symbol in *)
(*       let symb = Mtt.sigma_of_ord mtt symbol in *)
(*       let _ = Mtt.add_sigma mtt' symb (Mtt.arity mtt symbol) in *)
(*       let new_symb = Mtt.ord_of_sigma mtt' symb in *)
(*       List.iter (fun (_,qs,qs') -> *)
(*          List.iter (fun q' -> *)
(*             match action with *)
(*                | Undefined -> () *)
(*                | Det rhs -> *)
(*                   let new_state = (q',stat) in *)
(*                   let _ = Mtt.add_state mtt' new_state in *)
(*                   let state' = (Mtt.ord_of_state mtt') new_state in *)
(*                   let _ = Mtt.set_parameters mtt' state' (Mtt.num_of_parameters mtt state) in *)
(*                   Mtt.add_rule mtt' state' new_symb (trans mtt mtt' qs rhs) *)
(*                | Non_det rhs_list -> *)
(*                   List.iter (fun rhs -> *)
(*                      let new_state = (q',stat) in *)
(*                      let _ = Mtt.add_state mtt' new_state in *)
(*                      let state' = (Mtt.ord_of_state mtt') new_state in *)
(*                      let _ = Mtt.set_parameters mtt' state' (Mtt.num_of_parameters mtt state) in *)
(*                      Mtt.add_rule mtt' state' new_symb (trans mtt mtt' qs rhs) *)
(*                   ) rhs_list *)
(*          ) qs' *)
(*       ) rules *)
(*    ) mtt;  *)
   mtt'


let intersection mtt fsta enc = 
   let mtt' = inter mtt (fsta,enc) in mtt'
