
{ 

(* header *) 
open Parser_tl
open Lexing

let inc_line lexbuf =
   let pos = lexbuf.lex_curr_p in
   let n = pos.pos_lnum in
   lexbuf.lex_curr_p <- {pos_lnum = n+1;
                         pos_fname = pos.pos_fname;
                         pos_bol =  pos.pos_cnum;
                         pos_cnum = pos.pos_cnum
                        }

let new_token lexbuf =
   lexbuf.lex_start_p <- lexbuf.lex_curr_p

let error lexbuf msg =
   let pos = lexbuf.lex_start_p in
   let file = pos.pos_fname in
   let line = pos.pos_lnum in
   let cnum = pos.pos_cnum in
   let bol = pos.pos_bol in
   let msg' = "line "^ 
              string_of_int line ^ ", character "^ 
              string_of_int (cnum-bol) ^":\n"^ 
              msg ^" \""^ 
              Lexing.lexeme lexbuf ^"\"\n" in
   raise (Error.Lexing_error msg')

}

let letter = ['a'-'z' 'A'-'Z']
let number = ['0'-'9']
let id = letter ('_'|letter|number)* ("'")?

rule token = parse
    [' ' '\t']       { token lexbuf }    (* skip blanks *)
  | ['\n']           { inc_line lexbuf; token lexbuf }
  | '{'              { new_token lexbuf; SETBEGIN }
  | '}'              { new_token lexbuf; SETEND }
  | '('              { new_token lexbuf; BEGIN }
  | ')'              { new_token lexbuf; END }
  | '-''>'           { new_token lexbuf; TO }
  | ','              { new_token lexbuf; COMMA }
  | ';'              { new_token lexbuf; ENDRULE }
  | '.'              { new_token lexbuf; COPY }
  | '"'              { new_token lexbuf; QUOTE }
  | '/'              { new_token lexbuf; SLASH }
  | id as lxm        { new_token lexbuf; ID (lxm) }
  | eof              { new_token lexbuf; EOF }
  | _                {
                       new_token lexbuf;
                       error lexbuf "Lexing error: Illegal token"
                     }

{ 

(* trailer *) 

}
