type symbol = int
type state = int

type rhs = (symbol, state) RightHandSide.rhs

type action =
   Undefined |
   Det of rhs | 
   Non_det of rhs list

type domain =
   None |
   Any |
   NoneOf of symbol list |
   OneOf of symbol list

type rule = symbol -> action
type rule_set = (state,(domain * rule)) Hashtbl.t
type parameter_table = (state,int) Hashtbl.t

type ('a,'b,'c) mtt = 
   {
      mutable q : 'a SymbolTable.symbol_table;
      mutable pars : parameter_table;
      mutable sigma : 'b SymbolTable.symbol_table;
      mutable delta : 'c SymbolTable.symbol_table;
      mutable init : state list;
      mutable rules : rule_set;
      mutable sigma_arity : (symbol,int) Simple_table.table
   }

val empty : (('a -> 'a -> bool) * ('a -> string) * (string -> 'a)) ->
            (('b -> 'b -> bool) * ('b -> string) * (string -> 'b)) ->
            (('c -> 'c -> bool) * ('c -> string) * (string -> 'c)) -> 
            ('a,'b,'c) mtt

val add_state : ('a,'b,'c) mtt -> 'a -> unit
val add_sigma : ('a,'b,'c) mtt -> 'b -> int -> unit
   (* the new version operates only with one single alphabet sigma *)
val add_delta : ('a,'b,'c) mtt -> 'b -> unit
val add_rule : ('a,'b,'c) mtt -> state -> symbol -> rhs -> unit

val arity : ('a,'b,'c) mtt -> symbol -> int

val get_action : ('a,'b,'c) mtt -> state -> symbol -> action
val get_all_actions : ('a,'b,'c) mtt -> state -> action list
val get_domain : ('a,'b,'c) mtt -> state -> domain

val iter : ((state * symbol * action) -> unit) -> ('a,'b,'c) mtt -> unit
val iter_states : ('a -> int -> unit) -> ('a,'b,'c) mtt -> unit
val iter_sigma : ('b -> int -> unit) -> ('a,'b,'c) mtt -> unit
   (* the new version operates only with one single alphabet sigma *)
val iter_delta : ('b -> int -> unit) -> ('a,'b,'c) mtt -> unit

val num_of_parameters : ('a,'b,'c) mtt -> state -> int
val set_parameters : ('a,'b,'c) mtt -> state -> int -> unit
val change_parameters : ('a,'b,'c) mtt -> state -> int -> unit

val ord_of_state : ('a,'b,'c) mtt -> 'a -> state
val ord_of_sigma : ('a,'b,'c) mtt -> 'b -> symbol
   (* the new version operates only with one single alphabet sigma *)
val ord_of_delta : ('a,'b,'c) mtt -> 'b -> symbol

val state_of_ord : ('a,'b,'c) mtt -> state -> 'a
val sigma_of_ord : ('a,'b,'c) mtt -> symbol -> 'b
   (* the new version operates only with one single alphabet sigma *)
val delta_of_ord : ('a,'b,'c) mtt -> symbol -> 'b

val get_init : ('a,'b,'c) mtt -> state list
val get_initial_state : ('a,'b,'c) mtt -> state list
val set_init : ('a,'b,'c) mtt -> state -> unit
val set_initial_state : ('a,'b,'c) mtt -> state -> unit

val state_eq : ('a,'b,'c) mtt -> ('a -> 'a -> bool)
val sigma_eq : ('a,'b,'c) mtt -> ('b -> 'b -> bool)
   (* the new version operates only with one single alphabet sigma *)
val delta_eq : ('a,'b,'c) mtt -> ('b -> 'b -> bool)

val state_to_string : ('a,'b,'c) mtt -> ('a -> string)
val sigma_to_string : ('a,'b,'c) mtt -> ('b -> string)
   (* the new version operates only with one single alphabet sigma *)
val delta_to_string : ('a,'b,'c) mtt -> ('b -> string)

val state_from_string : ('a,'b,'c) mtt -> (string -> 'a)
val sigma_from_string : ('a,'b,'c) mtt -> (string -> 'b)
   (* the new version operates only with one single alphabet sigma *)
val delta_from_string : ('a,'b,'c) mtt -> (string -> 'b)

val print_states : ('a,'b,'c) mtt -> unit
val print_sigma : ('a,'b,'c) mtt -> unit
val print_delta : ('a,'b,'c) mtt -> unit
