module T = SymbolTable
module R = RightHandSide
module H = Hashtbl

type symbol = int
type state = int

type rhs = (symbol, state) R.rhs

type action =
   Undefined |
   Det of rhs | 
   Non_det of rhs list

type rule = symbol -> action
type rule_set = (state,rule) H.t

type ('a,'b,'c) mtt = 
   {
      mutable q : 'a T.symbol_table;
      mutable sigma : 'b T.symbol_table;
      mutable delta : 'c T.symbol_table;
      mutable init : state list;
      mutable rules : rule_set
   }

let undef = function x -> Undefined

let empty (f1,g1) (f2,g2) (f3,g3) =
   {
      q = T.create (f1,g1);
      sigma = T.create (f2,g2);
      delta = T.create (f3,g3);
      init = [];
      rules = H.create 101
   }

let add_state mtt state = T.add_symbol mtt.q state

let add_sigma mtt symbol = T.add_symbol mtt.sigma symbol

let add_delta mtt symbol = T.add_symbol mtt.delta symbol

let add_rule mtt state symbol rhs = 
   let rule = 
      try
         (H.find mtt.rules state)
      with
         Not_found -> undef in
   let action = rule symbol in
   let new_ac = 
      match action with
         Undefined -> Det rhs |
         Det rhs' -> Non_det [rhs;rhs'] |
         Non_det l -> Non_det (rhs :: l) in
   let rule' = (fun x -> if x=symbol then new_ac else rule x)
   in
      H.replace mtt.rules state rule'

let get_action mtt state symbol =
   try 
      (H.find mtt.rules state) symbol 
   with 
      Not_found -> Undefined

let ord_of_state mtt state = T.get_index mtt.q state
let ord_of_sigma mtt sigma = T.get_index mtt.sigma sigma
let ord_of_delta mtt delta = T.get_index mtt.delta delta
let state_of_ord mtt ord = T.get_symbol mtt.q ord
let sigma_of_ord mtt ord = T.get_symbol mtt.sigma ord
let delta_of_ord mtt ord = T.get_symbol mtt.delta ord

let get_init mtt = mtt.init
let set_init mtt q0 = 
    let init = mtt.init in
    if not (List.mem q0 init) then mtt.init <- (q0 :: init)

let state_to_string mtt = T.to_string mtt.q
let sigma_to_string mtt = T.to_string mtt.sigma
let delta_to_string mtt = T.to_string mtt.delta

let print_states mtt = 
   T.iter (fun x y -> Printf.printf "%s -> %i " ((state_to_string mtt) x) y)
   mtt.q; Printf.printf "\n"
let print_sigma mtt = 
   T.iter (fun x y -> Printf.printf "%s -> %i " ((sigma_to_string mtt) x) y)
   mtt.sigma; Printf.printf "\n"
let print_delta mtt = 
   T.iter (fun x y -> Printf.printf "%s -> %i " ((delta_to_string mtt) x) y)
   mtt.delta; Printf.printf "\n"
