open RightHandSide

let sigma_to_string mtt s =
   (* disallowed symbols are represented by a leading '~';
    * if a rule is defined for any symbol (=0), a '.' is returned;
    * otherwise the represented symbol is returned *)
   if (s<0) then "~" ^ (Mtt.sigma_to_string mtt (Mtt.sigma_of_ord mtt (-1*s))) else
   if (s=0) then "." else
   Mtt.sigma_to_string mtt (Mtt.sigma_of_ord mtt s)

let delta_to_string mtt s =
   (*Mtt.delta_to_string mtt (Mtt.delta_of_ord mtt s)*)
   sigma_to_string mtt s

let rec rhs_to_string mtt = function
   Empty_rhs -> "nil" |
   Parameter i -> "y_"^(string_of_int i) |
   Symbol (s,k,succ) ->
      delta_to_string mtt s
      (*Mtt.delta_to_string mtt (Mtt.delta_of_ord mtt s)*)^
      "("^(rhs_list mtt succ)^")" |
   Call (q,s,k,succ) ->
      Mtt.state_to_string mtt (Mtt.state_of_ord mtt q)^
      "(x_"^(string_of_int s)^
      (if k>0 then ","^(rhs_list mtt succ) else "")^")"
and rhs_list mtt = function
   [] -> "" |
   r::[] -> rhs_to_string mtt r |
   r::rs -> (rhs_to_string mtt r)^","^(rhs_list mtt rs)
and rhs_list' mtt = function
   [] -> "" |
   r::[] -> rhs_to_string mtt r |
   r::rs -> (rhs_to_string mtt r)^"; "^(rhs_list' mtt rs)

let rec pars_to_string n = function
   0 -> "" |
   1 -> "y_" ^ (string_of_int n) |
   i -> "y_" ^ (string_of_int (n-i+1)) ^ "," ^ (pars_to_string n (i-1))

let action_to_string mtt = function
   Mtt.Undefined -> "undef" |
   Mtt.Det ac -> rhs_to_string mtt ac |
   Mtt.Non_det acs -> "["^(rhs_list' mtt acs)^"]"

let state_to_string mtt q =
   Mtt.state_to_string mtt (Mtt.state_of_ord mtt q)

let rule_to_string mtt (q,s,a) =
   let state = state_to_string mtt q in
   let sym = sigma_to_string mtt s 
             (*Mtt.sigma_to_string mtt (Mtt.sigma_of_ord mtt s)*) in
   let n = Mtt.num_of_parameters mtt q in
   let pars = pars_to_string n n in
   let act = action_to_string mtt a in
   if n = 0 then
      state ^"("^ sym ^") -> "^ act ^"\n"
   else
      state ^"("^ sym ^","^ pars ^") -> "^ act ^"\n"

let to_string mtt =
   let str = ref "" in
   Mtt.iter (fun (q,s,a) ->
      let state = state_to_string mtt q in
      let sym = sigma_to_string mtt s in
      let n = Mtt.num_of_parameters mtt q in
      let pars = pars_to_string n n in
      let act = action_to_string mtt a in
      if n = 0 then
         str := !str ^ (state^"("^sym^") -> "^act^"\n")
      else
         str := !str ^ (state^"("^sym^","^pars^") -> "^act^"\n")
   ) mtt; !str

let print mtt = 
   Mtt.iter (fun (q,s,a) ->
      let state = state_to_string mtt q in
      let sym = Mtt.sigma_to_string mtt (Mtt.sigma_of_ord mtt s) in
      let n = Mtt.num_of_parameters mtt q in
      let pars = pars_to_string n n in
      let act = action_to_string mtt a in
      if n = 0 then
         Printf.printf "%s(%s) -> %s\n" state sym act
      else
         Printf.printf "%s(%s,%s) -> %s\n" state sym pars act
   ) mtt
