
/* File parser_tl.mly */

%{
   
open Lexing

let count = ref 0
let reset_count () = count := 0

let pars = ref []
let reset_pars () = pars := []
let add_par p = pars := (p :: !pars)
let get_pars () = !pars
let is_par p = List.mem p !pars

   
let syntax_error msg =
   let pos = Parsing.symbol_start_pos () in
      (* unfortunately, pos.pos_fname is empty *)
   let file = pos.pos_fname in
   let line = pos.pos_lnum in
   let cnum = pos.pos_cnum in
   let bol = pos.pos_bol in
   let msg' = "line "^ 
              string_of_int line ^", character "^ 
              string_of_int (cnum-bol) ^"\n Syntax Error: "^ 
              msg ^"\n" in
   raise (Error.Syntax_error msg')

let parse_error msg =
   let pos = Parsing.symbol_start_pos () in
   let line = pos.pos_lnum in
   let cnum = pos.pos_cnum in
   let bol = pos.pos_bol in
   let msg' = "line "^ 
              string_of_int line ^", character "^ 
              string_of_int (cnum-bol) ^"\n"^
              msg ^"\n" in
   raise (Error.Parse_error msg')

let warning msg =
   let pos = Parsing.symbol_start_pos () in
   let line = pos.pos_lnum in
   let cnum = pos.pos_cnum in
   let bol = pos.pos_bol in
   let msg' = "Warning: line "^ 
              string_of_int line ^", character "^ 
              string_of_int (cnum-bol) ^": "^
              msg ^"\n" in
   print_string msg'; flush stdout

%}
%token <string> ID
%token SETBEGIN SETEND
%token BEGIN END
%token COMMA
%token TO
%token QUOTE
%token SLASH
%token ENDRULE
%token COPY
%token EOL
%token EOF
%left COMMA
%start prog
%type <Tl_types.program> prog
%%
prog:
        rules EOF                  { let _ = reset_count () in $1 }
;
rules:
        rule ENDRULE rules         { $1 :: $3 }
      | rule ENDRULE               { $1 :: [] }
;
rule:
        ID BEGIN head END TO rhs   { let (pat,pars) = $3 in (($1,pat,pars),$6) }
;
head:
        QUOTE ID QUOTE COMMA pars  { ($2,$5) }
      | QUOTE ID QUOTE             { ($2,[]) }
;
pars:
        par COMMA pars             { let _ = add_par $1 in $1 :: $3 }
      | par                        { let _ = add_par $1 in $1 :: [] }
;
rhs:
        ID                                      { 
                                                   if (is_par $1) 
                                                   then (Tl_types.Param $1)
                                                   else (Tl_types.Out ($1,[]))
                                                }
      | ID BEGIN END                            { Tl_types.Out ($1,[]) }
      | ID BEGIN QUOTE pat QUOTE END            { Tl_types.Fun ($1,$4,[]) }
      | ID BEGIN QUOTE pat QUOTE COMMA rhss END { Tl_types.Fun ($1,$4,$7) }
      | ID BEGIN rhss END                       { Tl_types.Out ($1,$3) }
      | COPY                                    { Tl_types.Copy }
;
rhss:
        rhs COMMA rhss             { $1 :: $3 }
      | rhs                        { $1 :: [] }
;
pat:
        ID SLASH pat               { $1 :: $3 }
      | ID                         { $1 :: [] }
;
par:
        ID                         { $1 }
;
        

