exception Error
open ParserTypes
open RightHandSide
open Mtt

module H = Hashtbl

let rec rhs_to_string mtt = function
   Empty_rhs -> "" |
   Parameter i -> "y_"^(string_of_int i) |
   Symbol (s,k,succ) ->
      if (s=0) then 
         ".(" ^ (rhs_list mtt succ) ^ ")"
      else
         delta_to_string mtt (delta_of_ord mtt s)^
         "("^(rhs_list mtt succ)^")" |
   Call (q,s,k,succ) ->
      state_to_string mtt (state_of_ord mtt q)^
      "(x_"^(string_of_int s)^
      (if k>0 then ","^(rhs_list mtt succ) else "")^")"
and rhs_list mtt = function
   [] -> "" |
   r::[] -> rhs_to_string mtt r |
   r::rs -> (rhs_to_string mtt r)^","^(rhs_list mtt rs)

let eq x y = 
   String.compare x y == 0

let to_str x = x

let from_str x = x

let get_state_num mtt s =
   try
      ord_of_state mtt s
   with
      Not_found -> (add_state mtt s; ord_of_state mtt s)

let get_delta_num mtt s =
   try
      ord_of_delta mtt s
   with
      Not_found -> (add_delta mtt s; ord_of_delta mtt s)

let rec mk_rhs mtt vars pars = function
   Sym s -> 
      if H.mem pars s then
         Parameter (H.find pars s)
      else
         Symbol (get_delta_num mtt s,0,[]) |
   App (s,l) ->
      (match l with
         [] -> 
            Symbol (get_delta_num mtt s,0,[]) |
         x::ll ->
            match x with
               Sym s' ->
                  if H.mem vars s' then
                     let stat = get_state_num mtt s
                     in
                        Call (stat,H.find vars s', List.length ll, 
                              List.map (mk_rhs mtt vars pars) ll)
                  else
                     Symbol (get_delta_num mtt s,List.length l, 
                             List.map (mk_rhs mtt vars pars) l) |
               _ -> Symbol (get_delta_num mtt s,List.length l, 
                            List.map (mk_rhs mtt vars pars) l) ) |
   Any_rhs l ->
      (match l with
         [] -> Symbol (0,0,[]) |
         _  -> Symbol (0,List.length l,List.map (mk_rhs mtt vars pars) l) )

let rec mk_rhs2 mtt (var,sym) vars pars = function
   Sym s -> 
      if H.mem pars s then
         Parameter (H.find pars s)
      else
         let symb = get_delta_num mtt s in
            if symb == var then
               Symbol (sym,0,[])
            else
               Symbol (symb,0,[]) |
   Any_rhs l ->
      (match l with
         [] -> Symbol (0,0,[]) |
         _  -> Symbol (0,List.length l,List.map (mk_rhs2 mtt (var,sym) vars pars) l) ) |
   App (s,l) ->
      match l with
         [] -> 
            let symb = get_delta_num mtt s in
               if symb == var then
                  Symbol (sym,0,[])
               else
                  Symbol (symb,0,[]) |
         x::ll ->
            match x with
               Sym s' ->
                  if H.mem vars s' then
                     let stat = get_state_num mtt s
                     in
                        Call (stat,H.find vars s', List.length ll, 
                              List.map (mk_rhs2 mtt (var,sym) vars pars) ll)
                  else
                     let symb = get_delta_num mtt s in
                        if symb == var then
                           Symbol (sym,List.length l,
                              List.map (mk_rhs2 mtt (var,sym) vars pars) l)
                        else
                           Symbol (symb,List.length l,
                              List.map (mk_rhs2 mtt (var,sym) vars pars) l) |
               _ ->
                  let symb = get_delta_num mtt s in
                     if symb == var then
                        Symbol (sym,List.length l,
                           List.map (mk_rhs2 mtt (var,sym) vars pars) l)
                     else
                        Symbol (symb,List.length l,
                           List.map (mk_rhs2 mtt (var,sym) vars pars) l)

let read debug file =
   let first = ref true in
   let rule_num = ref 0 in
   let mtt = empty (eq,to_str,from_str) (eq,to_str,from_str) (eq,to_str,from_str) in
   let in_channel = 
      try
         open_in file
      with
         Sys_error e -> raise Error 
   in
      (try 
         let lexbuf = Lexing.from_channel in_channel in
            while true do
               let (state,sym,vars,pars,rhs,cond,str) = 
                  Parser.main Lexer.token lexbuf in
(*                let _ = *)
(*                   Printf.printf "Found Variables:\n"; *)
(*                   List.iter (fun x -> Printf.printf "%s " x) vars; *)
(*                   Printf.printf "\n" in *)
(*                let _ = *)
(*                   Printf.printf "Found Parameters:\n"; *)
(*                   List.iter (fun x -> Printf.printf "%s " x) pars; *)
(*                   Printf.printf "\n" in *)
               let v =
                  let h = Hashtbl.create 101 in
                  let pos = ref 0 in
                     List.iter (fun x -> (pos:=!pos+1; H.add h x !pos)) vars;
                     h in
               let p =
                  let h = Hashtbl.create 101 in
                  let pos = ref 0 in
                     List.iter (fun x -> (pos:=!pos+1; H.add h x !pos)) pars;
                     h in
               let rhs' = mk_rhs mtt v p rhs in
               let q = get_state_num mtt state in
               let _ = set_parameters mtt q (List.length pars) in
               let _ =
                  if !first then
                     (set_init mtt q; first:=false) in
               let arity = List.length vars in
               let _ =
                  match cond with
                     Empty_cond -> 
                        (match sym with
                           | Expl_sym s' -> add_sigma mtt s' arity
                           | Any_sym -> ()
                           | Excl_sym s' -> 
                                 if debug then Printf.printf "symbol >%s< found\n" s';
                                 add_sigma mtt s' arity) |
                     Opt (s,l) -> List.iter (fun sym ->
                        (*
                        match sym with
                           | Expl_sym s' -> add_sigma mtt s' arity
                           | Any_sym -> ()
                           | Excl_sym s' -> add_sigma mtt s' arity
                        *)
                        add_sigma mtt sym arity
                     ) l |
                     Excl (s,l) -> () in
               let _ =
                  match cond with
                     Empty_cond -> 
                        let s = match sym with
                           | Expl_sym s' -> ord_of_sigma mtt s'
                           | Any_sym -> 0
                           | Excl_sym s' -> 
                                 if debug then Printf.printf "symbol >- %i< found\n" (ord_of_sigma mtt s');
                                 -1 * (ord_of_sigma mtt s') in
                        add_rule mtt (ord_of_state mtt state) s rhs' |
                     Opt (s,l) -> 
                        List.iter 
                           (fun x -> 
                              let rhs' = mk_rhs2 mtt (get_delta_num mtt s,
                                         get_delta_num mtt x) v p rhs in
                              add_rule mtt 
                                 (ord_of_state mtt state) 
                                 (ord_of_sigma mtt x) rhs'
                           ) l |
                     Excl (s,l) -> ()
               in
                  rule_num:=!rule_num + 1 ;
                  if debug then Printf.printf "%i. rule:     %s \n" !rule_num str;
                  if debug then Printf.printf "rec. as:      %s\n" (rhs_to_string mtt rhs');
                  flush stdout
            done
      with 
         Lexer.Eof -> ());
      mtt
