exception Error

open Tl_types
open RightHandSide
open Mtt

let eq x y     = String.compare x y == 0
let to_str x   = x
let from_str x = x

let count = ref 0
let reset_count () = count := 0
let next () = count := !count+1
let get () = !count

let first = ref true
let init_fun = ref 0

let to_mtt prog =
   let m = empty (eq,to_str,from_str) (eq,to_str,from_str) (eq,to_str,from_str)
   in let dummy = try ord_of_sigma m "." 
                  with Not_found -> (add_sigma m "." 2; ord_of_sigma m ".")
   in let leaf  = try ord_of_sigma m "<empty>" 
                  with Not_found -> (add_sigma m "<empty>" 2; ord_of_sigma m "<empty>")
   in let todo  = ref []

   in let mk_fun f i j s = "_"^ f ^"_"^ (string_of_int i) ^"_"^ (string_of_int j) ^"_"^ s

   in let rec mk_par_list l = function
      | 0 -> l 
      | i -> mk_par_list ((Parameter i) :: l) (i-1)

   in let rec do_pattern f rank g i = function
      | s::t::[] -> 
            let sym = try ord_of_sigma m s
                      with Not_found -> (add_sigma m s 2; ord_of_sigma m s)
            in let f' = try ord_of_state m f
                        with Not_found -> (add_state m f; ord_of_state m f)
            in let _ = set_parameters m f' rank
            in let p = if (i=1) then (Call (g,2,rank,mk_par_list [] (rank)))::[] else []
            in let rhs = Call (f',0,rank,(mk_par_list [] (rank-1)) @ p)
            in todo := (g,leaf,Symbol(leaf,0,[])) :: (g,sym,rhs) :: !todo
      | s::r ->
            let sym = try ord_of_sigma m s
                      with Not_found -> (add_sigma m s 2; ord_of_sigma m s)
            in let f_sym = mk_fun f (get()) i s
            in let f' = try ord_of_state m f_sym
                        with Not_found -> (add_state m f_sym; ord_of_state m f_sym)
            in let _ = set_parameters m f' rank
            in let p = if (i=1) then (Call (g,2,rank,mk_par_list [] rank))::[] else []
            in let rhs = Call (f',0,rank,(mk_par_list [] (rank-1)) @ p)
            in (todo := (g,leaf,Symbol(leaf,0,[])) :: (g,sym,rhs) :: !todo; do_pattern f rank f' (i+1) r)
      | _ -> ()

   in let rec do_rhs body par fun_sym k sym = 
      match body with
         | [] -> Symbol (leaf,0,[])
         | b::bs -> 
               match b with
                  | Out (s,args) -> 
                        let argn = List.length args
                        in let rank = 2
                        in let s' = try ord_of_sigma m s 
                                    with Not_found -> (add_sigma m s rank; ord_of_sigma m s)
                        in let cont = do_rhs args par fun_sym k sym
                        in let next = do_rhs bs par fun_sym k sym
                        in Symbol (s',rank, cont :: next  :: [])
                  | Param y -> 
                        let j = 
                           let rec f i = function
                              | x::xs -> if (eq x y) then i else f (i+1) xs
                              | [] -> 0
                           in f 1 par
                        in Parameter j
                  | Fun (f,pat,pars) ->
                        if (List.length pat)>1 then
                           let f_sym = mk_fun f (get()) 1 (List.hd pat)
                           in let _ = next()
                           in let rank = (List.length pars) + 1
                           in let f' = try ord_of_state m f_sym
                                       with Not_found -> (add_state m f_sym; ord_of_state m f_sym)
                           in let _ = set_parameters m f' rank
                           in let _ = do_pattern f rank f' 1 pat
                           in let pars' = List.map (fun r -> do_rhs [r] par fun_sym k sym) pars
                           in let last_par = (do_rhs bs par fun_sym k sym) :: []
                           in let _ = todo := (f',dummy,(Call (f',2,rank,mk_par_list [] rank))) :: !todo
                           in Call (f',0,rank,pars' @ last_par)
                        else
                           let rank = (List.length pars) + 1
                           in let f' = try ord_of_state m f
                                       with Not_found -> (add_state m f; ord_of_state m f)
                           in let _ = set_parameters m f' rank
                           in let pars' = List.map (fun r -> do_rhs [r] par fun_sym k sym) pars
                           in let last_par = (do_rhs bs par fun_sym k sym) :: []
                           in let _ = todo := (f',dummy,(Call (f',2,rank,pars' @ (Parameter rank :: [])))) :: !todo
                           in Call (f',0,rank,pars' @ last_par)
                  | Copy -> 
                        let _ = add_rule m fun_sym dummy (Symbol (dummy,2,Call (fun_sym,1,0,[]) :: Call (fun_sym,2,0,[]) :: []))
                        in let _ = add_rule m fun_sym leaf (Symbol (leaf,0,[]))
                        in Symbol (sym,2,Call (fun_sym,1,0,[]) :: Call (fun_sym,2,0,[]) :: [])

   in let mk_rule ((f,s,pars),body) =
      let fun_sym = try ord_of_state m f
                    with Not_found -> (add_state m f; ord_of_state m f)
      in let rank = (List.length pars) + 1
      in let _    = set_parameters m fun_sym rank
      in let _    = if !first then (init_fun := fun_sym; first := false)
      in let sym  = (add_sigma m s 2; ord_of_sigma m s)
      in let rhs  = do_rhs (body :: []) pars fun_sym rank sym
      in (add_rule m fun_sym sym rhs; 
          add_rule m fun_sym leaf (Parameter rank);
          add_rule m fun_sym dummy (Call (fun_sym,2,rank,mk_par_list [] rank)))

   in let rec doit = function
      | [] -> ()
      | r::p -> (mk_rule r; doit p)

   in (doit prog; 
       List.iter (fun (a,b,c) -> add_rule m a b c) !todo; 
       let fun_sym = (add_state m "_initial_function"; ord_of_state m "_initial_function")
       in let _ = set_parameters m fun_sym 0
       in let _ = set_init m fun_sym
       in let f = !init_fun
       in let k = num_of_parameters m f
       in let rhs = (Call (f,0,k,(let rec doit = function 
                                     | 0 -> [] 
                                     | i -> Symbol (leaf,0,[]) :: (doit (i-1)) 
                                  in doit k)))
       in add_rule m fun_sym dummy rhs;
       m)

let read debug file =
   let in_channel = open_in file
   in let lexbuf = Lexing.from_channel in_channel
   in let program = Parser_tl.prog Lexer_tl.token lexbuf
(*    in let _ = Printf.printf "%s\n" (to_string program) *)
   in let mtt = to_mtt program
(*    in let _ = Printf.printf "%s\n" (Mtt_to_string.to_string mtt) *)
   in mtt

