module type Simple_table =
   sig
      exception First_entry_exists
      type ('a,'b) table
      val create : unit -> ('a,'b) table
      val add : ('a,'b) table -> 'a * 'b -> unit
      val rm : ('a,'b) table -> 'a -> unit
      val replace : ('a,'b) table -> 'a -> 'b -> unit
      val mem : ('a,'b) table -> 'a -> bool
      val lookup : ('a,'b) table -> 'a -> 'b
      val iter : ('a -> 'b -> unit) -> ('a,'b) table -> unit
      val fold : ('a -> 'b -> 'c -> 'c) -> ('a,'b) table -> 'c -> 'c
      val to_list : ('a,'b) table -> 'a list
      val to_pair_list : ('a,'b) table -> ('a * 'b) list
      val size : ('a,'b) table -> int
   end

module H = Hashtbl

exception First_entry_exists

type ('a,'b) table =
   {
      mutable atob : ('a,'b) H.t;
      mutable size : int
   }

let create () =
   {
      atob = H.create 101;
      size = 0
   }

let inc_size table = table.size <- table.size + 1

let dec_size table = table.size <- table.size - 1

let add table (a,b) =
   let a' = H.mem table.atob a
   in
      if not a' then
         (H.add table.atob a b; inc_size table)
      else
         raise First_entry_exists

let rm table a = 
   if H.mem table.atob a then
      (H.remove table.atob a; dec_size table)

let replace table a b = H.replace table.atob a b

let mem table a = H.mem table.atob a

let lookup table a = H.find table.atob a

let iter f table = H.iter f table.atob

let fold f table init = H.fold f table.atob init

let to_list table = H.fold (fun a _ r -> a :: r) table.atob []

let to_pair_list table = H.fold (fun a b r -> (a,b) :: r) table.atob []

let size table = table.size
