type item = int

module IntItem =
   struct
      type t = item
      let eq x y = x = y
      let to_string = string_of_int
   end

module T = Basic_table.Make (IntItem)

type 'a symbol_table =
   {
      mutable tbl : 'a T.table;
      mutable last : int
   }

let create (eq,to_str,from_str) = 
   {
      tbl = T.create (eq,to_str,from_str);
      last = 0 (* i.e. the first new symbol gets index 1 *)
   }

let mem tbl symbol = T.mem tbl.tbl symbol

let add tbl symbol =
   if not (mem tbl symbol) then
      let index = tbl.last + 1
      in
         T.add tbl.tbl symbol index; tbl.last <- index; index
   else
      T.lookup tbl.tbl symbol

let add_symbol tbl symbol =
   if not (mem tbl symbol) then
      let index = tbl.last + 1
      in
         T.add tbl.tbl symbol index;
         tbl.last <- index

let rm_symbol tbl symbol =
   T.rm tbl.tbl symbol;
   tbl.last <- tbl.last - 1

let last_index tbl = tbl.last

let lookup tbl symbol = T.lookup tbl.tbl symbol

let get_index tbl symbol = T.lookup tbl.tbl symbol

let get_symbol tbl index = List.hd (T.rev_lookup tbl.tbl index) 

let iter f tbl = T.iter f tbl.tbl

let eq tbl = T.eq tbl.tbl

let to_string tbl = T.to_string tbl.tbl

let from_string tbl = T.from_string tbl.tbl
