type fun_sym = string
type symbol  = string
type par     = string
type pattern = string list

type head    = fun_sym * symbol * par list
type body    = Out    of string * body list 
             | Fun    of string * pattern * body list
             | Param  of string  
             | Copy

type rule    = head * body
type program = rule list

let rec pat_to_string = function
   | [] -> ""
   | p::[] -> p
   | p::ps -> p ^"/"^ (pat_to_string ps)

let rec rhs_to_string body =
   match body with
      | Out(s,args)    -> s ^"("^ (args_to_string args) ^")"
      | Fun(f,p,pars)  -> f ^"(["^ (pat_to_string p) ^"]"^ (match pars with []->"" | _->",") ^ (args_to_string pars) ^")"
      | Param(p)       -> p
      | Copy           -> "<copy>"
and args_to_string = function
   | [] -> ""
   | a::[] -> (rhs_to_string a)
   | a::r  -> (rhs_to_string a) ^","^ (args_to_string r)

let rec pars_to_string = function
   | [] -> ""
   | p::[] -> p
   | p::ps -> p ^","^ (pars_to_string ps)

let head_to_string (f,pat,pars) =
   f ^"("^ pat ^ (match pars with [] -> "" | _ -> ","^ (pars_to_string pars)) ^")"

let rec rule_to_string (head,body) =
   (head_to_string head) ^" --> "^ (rhs_to_string body)

let rec to_string prog = 
   match prog with
      | r::p -> (rule_to_string r) ^"\n"^ (to_string p)
      | []   -> "\n"

