module T = SymbolTable

type elt = Sym.t
type t = elt T.symbol_table

let empty = 
   let eq s1 s2 = (String.compare s1 s2) = 0 in
   T.create (eq, Sym.to_string, Sym.from_string)

let mem a alph = T.mem alph a

let add a alph = 
   let alph' = alph in
   T.add_symbol alph' a; alph'

let remove a alph =
   let alph' = alph in
   T.rm_symbol alph' a; alph'

let iter f alph = T.iter (fun item _ -> f item) alph

let cardinal alph =
   let c = ref 0 in
   T.iter (fun item _ -> c := !c + 1) alph; !c

let subset alph1 alph2 =
   let b = ref true in
   T.iter (fun item _ -> b := !b && (T.mem alph2 item)) alph1; !b
let equal alph1 alph2 = (subset alph1 alph2) && (subset alph2 alph1)

let to_string alph =
   let to_str = T.to_string alph
   and s = ref "" in
   T.iter (fun item _ -> s := !s ^" "^ (to_str item)) alph; !s

let print_alphabet alph = print_string (to_string alph)

let from_list l =
   let alph' = empty in
   List.iter (fun s -> T.add_symbol alph' s) l; alph'

let fold f alph init =
   let r = ref init in
   T.iter (fun item _ -> r := (f item !r)) alph; !r
