module Symb = Symbol.Symbol
module Alpha = Alphabet.Alphabet
module Terms = Term.Term
module Stat = State.State
module Stats = States.States
module Rule = Rules.Rules
module Autom = Automaton.Automaton

type symbol = Symbol.Symbol.t
type alphabet = Alphabet.Alphabet.t
type term = Term.Term.t
type state = State.State.t
type states = States.States.t
type rules = Rules.Rules.t
type automaton = (alphabet,states,rules) Automaton.automata_type
type tree = symbol Tree.t

exception Alphabets_not_equal = Autom.Alphabets_not_equal
exception Empty_tree


let debug = Autom.debug
let set_debug = Autom.set_debug
let empty = Autom.empty
let membership = Autom.membership
let to_string = Autom.to_string
let determinize = Autom.determinize
let complement = Autom.complement
let union = Autom.union
let inter = Autom.inter
let emptyness = Autom.emptyness
let from_list = Autom.from_list

let alph_iter f automaton = Alpha.iter f automaton.Automaton.alphabet

let sym_from_string = Symb.from_string
let sym_to_string = Symb.to_string
let sym_compare = Symb.compare
let empty_tree = "-23"

let state_from_string = Stat.from_string
let state_to_string = Stat.to_string
let state_compare = Stat.compare

let states_from_list = Stats.from_list
let states_to_string = Stats.to_string
let states_compare = Stats.compare
let states_fold = Stats.fold
let final_states automaton = 
   Stats.fold (fun q result -> 
      (int_of_string (Stat.to_string q)) :: result
   ) automaton.Automaton.final_states []

let next_states automaton = Rule.find automaton.Automaton.rules
let iter_rules f automaton = Rule.iter3 f automaton.Automaton.rules
let add_rule automaton symbol args states = 
   (* 7 is the expected number of rules for this symbol; it is
      needed by Rules.Rules.add for initializing a hash table *) 
   Rule.add automaton.Automaton.rules symbol args states 7

let tree_to_term t =
   let module T = Tree in
   let empty = Term.L (Symb.from_string "-23") in
   let rec convert_tree = function
      T.Empty_tree -> empty |
      T.Node (label,sons) -> Term.N (label, convert_forest sons, empty)
   and convert_forest = function 
      [] -> empty |
      t::f -> Term.N (T.label t, convert_forest (T.sons t), convert_forest f)
   in
      convert_tree t

let tree_to_term' t enc =
   let module T = Tree in
   let empty = Term.L (Symb.from_string "-23") in
   let rec convert_tree = function
      T.Empty_tree -> empty |
      T.Node (label,sons) ->
         let label' = enc label in
         Term.N (label', convert_forest sons, empty)
   and convert_forest = function
      [] -> empty |
      t::f ->
         let label' = enc (T.label t) in
         Term.N (label', convert_forest (T.sons t), convert_forest f)
   in
      convert_tree t

let mem automaton t = membership automaton (tree_to_term t)
        
let is_deterministic automaton = 
   Rule.is_deterministic automaton.Automaton.rules
