type 'a t = Empty_tree | Node of 'a * 'a t list

exception Empty_forest

let sons = function
   Empty_tree -> [] |
   Node(_,l) -> l

let nth i t = List.nth (sons t) i

let label = function
   Empty_tree -> raise Empty_forest |
   Node(n,_) -> n

let rec indent_to_string str = function
   [] -> str |
   x::xs ->
      if x > 0 then
         let s =
            let r = ref "" in
               (for i=0 to x do r := (!r) ^ " " done; (!r))
            in
               indent_to_string (str ^ s ^ "|") xs
         else
            indent_to_string (str) xs
   
let rec tree_to_xml f ind = function
   Empty_tree -> "" |
   Node(lab,succ) ->
      let tag = f lab in
      let indent = 
         let r = ref "" in
            (for i=0 to ind do r := (!r) ^ "   " done; (!r)) in
      let s_tag = "<"^tag^">" in
      let cont = forest_to_xml f (ind+1) succ in
      let e_tag = "</"^tag^">"in
         indent ^ s_tag ^ "\n" ^
         cont ^
         indent ^ e_tag ^ "\n"
and forest_to_xml f ind = function
   [Empty_tree] -> "" |
   [] -> "" |
   x::xs -> (tree_to_xml f ind x) ^ (forest_to_xml f ind xs)
               
let rec tree_to_string f ind = function
   Empty_tree -> 
      (indent_to_string "" (List.rev ind)) ^
      "- []\n" |
   Node(lab,succ) -> 
      let ind' = (String.length (f lab)) + 4
      in
         (indent_to_string "" (List.rev ind)) ^
         "- " ^ (f lab) ^ " --+\n" ^ 
         (forest_to_string f (ind'::ind) "" succ) ^ 
         (indent_to_string "" (List.rev ind)) ^ "\n" ^
         (indent_to_string "" (List.rev ind)) ^ "\n"
and forest_to_string f ind str = function
   [Empty_tree] ->
      str ^ (indent_to_string "" (List.rev ind)) ^
      "- []\n" |
   [] -> 
      str ^ (indent_to_string "" (List.rev ind)) ^
      "- []\n" |
   x::xs ->  forest_to_string f ind (str ^ (tree_to_string f ind x)) xs

let to_string x f = tree_to_string f [0] x

let to_xml x f = tree_to_xml f 0 x

