let list_to_string f op cl sep l = 
   let rec g = function
      [] -> "" |
      e::[] -> f e |
      e::es -> (f e) ^ sep ^ (g es)
   in
      op ^ (g l) ^ cl

   (* get_jth_of_list returns all those elements l[i] of a list l
      such that i is element of a given list of positions. If the
      list is returned it is reversed in order to keep the 
      original order of the elements. *)
let rec get_jth_of_list ls pos_list = 
   let rec get result i l = function
      [] -> List.rev result |
      j::js ->
         match l with
            [] -> result |
            x::xs ->
               if j=i then
                  get (x::result) (i+1) xs js
               else
                  get result (i+1) xs (j::js)
   in
      get [] 1 ls pos_list



   (* tuple lol with lol=l1,...ln returns the list of all possible lists 
      l'1,...,l'k such that each l'i (1 <= i <= k) consists of elements
      e1,...,en (n is the length of lol) where ej is element of lj for
      all j=1,...,n *)
let rec single = function
   | [] -> []
   | e::es -> [e] :: single es

let rec prepend x = function
   | [] -> []
   | l::ls -> (x::l) :: prepend x ls

let rec pair lst = function
   | [] -> [] 
   | e::es -> (prepend e lst) @ pair lst es

let rec prepend' result l = function
   | [] -> result
   | e::es -> prepend' ((e::l) :: result) l es

let rec combine result lst = function
   | [] -> result
   | e::es -> combine (prepend' result [e] lst) lst es

let tuple lst = 
   let rec doit = function
      | [] -> []
      | lst::[] -> single lst
      | lst1::lst2::[] -> combine [] lst1 lst2
      | lst::lsts -> pair (doit lsts) lst
   in List.map List.rev (doit (List.rev lst))



let rec powerset = function
   | [] -> [[]] 
   | x :: xs -> let tmp = powerset xs in tmp @ (List.map (fun e -> x :: e) tmp)

