type 'a t = 
   { mutable count : int;
     mutable head : int;
     mutable last : int;
     mutable contained : ('a,int) Hashtbl.t;
     mutable wlist : (int,'a) Hashtbl.t }

exception Empty

let init = 101

let empty () = 
   { count = 0;
     head = 0;
     last = 0;
     contained = Hashtbl.create init;
     wlist = Hashtbl.create init }

let clear w =
   w.count <- 0;
   w.head <- 0;
   w.last <- 0;
   Hashtbl.clear w.contained;
   Hashtbl.clear w.wlist

let is_empty w = 
   w.count == 0 

let choose w = 
   if w.count == 0 then raise Empty
   else Hashtbl.find w.wlist (w.head) 

let rec cleanup w i =
   if i == w.last then
      w.last <- w.last - 1
   else 
      let x = Hashtbl.find w.wlist (i+1) in
         Hashtbl.add w.wlist i x;
         Hashtbl.add w.contained x i;
         Hashtbl.remove w.wlist (i+1);
         Hashtbl.remove w.contained x;
         cleanup w (i+1)

let remove w x =
   if w.count == 0 then raise Empty
   else
      let i = Hashtbl.find w.contained x in
         w.count <- w.count - 1;
         Hashtbl.remove w.contained x;
         Hashtbl.remove w.wlist i;
         if w.head == w.last then
            (w.head <- 0; w.last <- 0)
         else
            if i == w.head then
               w.head <- w.head + 1
            else
               cleanup w i

let single x =
   let w = empty () in
      w.count <- 1;
      w.head <- w.count;
      w.last <- w.count;
      Hashtbl.add w.contained x w.count;
      Hashtbl.add w.wlist w.count x; w

let extract w =
   if is_empty w then
      raise Empty
   else
      let x = Hashtbl.find w.wlist w.head in
         w.count <- w.count - 1;
         Hashtbl.remove w.contained (Hashtbl.find w.wlist w.head);
         Hashtbl.remove w.wlist w.head;
         if w.head == w.last then
            (w.head <- 0; w.last <- 0)
         else
            w.head <- w.head + 1;
         x


let mem w = Hashtbl.mem w.contained

let _add w x =
   if not (mem w x) then
      (if w.count == 0 then w.head <- 1;
      w.count <- w.count + 1;
      w.last <- w.last + 1;
      Hashtbl.add w.wlist w.last x;
      Hashtbl.add w.contained x w.last)

let rec add w = function
   [] -> () |
   x::xs -> _add w x; add w xs
   
let union w1 w2 =
   Hashtbl.iter (fun i x -> _add w1 x) w2.wlist
